from copy import deepcopy
import pandas as pd
import networkx as nx
from itertools import chain
from hashlib import sha1
from collections import defaultdict, Counter
from typing import List, Dict
import logging

logger = logging.getLogger(__name__)


def conditional_add_edge(u, v, graph: nx.Graph, attributes: dict) -> None:
    if graph.has_edge(u, v):
        # update edge
        cur_edge_data = graph.get_edge_data(u, v)
        for key, value in attributes.items():
            if key in ["time"]:
                pass
            else:
                cur_edge_data[key] += value
    else:
        graph.add_edge(u, v, **attributes)


def update_edge_details(dict1: dict, dict2: dict) -> dict:
    new_dict = {}
    for k, v in dict2.items():
        if k in ["pos", "time", "data"]:
            continue
        if k not in dict1:
            new_dict[k] = v
        else:
            new_dict[k] = dict1[k] + v
    return new_dict


def update_node_details(final_node_data: dict, prev_node_data: dict) -> dict:
    for k, v in prev_node_data.items():
        if k in ["pos", "time", "data"]:
            continue
        if k not in final_node_data:
            final_node_data[k] = v
        else:
            final_node_data[k] += v

    return final_node_data


def hash_event(e: list) -> str:
    return sha1(str(e).encode()).hexdigest()[:8]


class Prediction(dict):
    """Data Structure to wrap a Prediction dictionary with additional functionality

    Provides interfaces for generating NetworkX graphs from prediction information

    """

    def __init__(self, prediction: dict = None, metadata: list = None, **kwargs):
        """Constuct Prediction class object. Optionally provide
        metadata information for sequence

        Args:
            prediction (dict, optional): Prediction dictionary. Defaults to None.
            metadata (list, optional): Optional sequence of metadata
                pertaining to prediction object (from model). Defaults to None.
        """
        if prediction is None:
            prediction = kwargs

        self._prediction = deepcopy(prediction)
        self.confidence: float = self._prediction["confidence"]
        self.confluence: float = self._prediction["confluence"]
        self.emotives: dict = self._prediction["emotives"]
        self.entropy: float = self._prediction["entropy"]
        self.evidence: float = self._prediction["evidence"]
        self.extras: list = self._prediction["extras"]
        self.fragmentation: float = self._prediction["fragmentation"]
        self.frequency: int = self._prediction["frequency"]
        self.future: list = self._prediction["future"]
        self.grand_hamiltonian: float = self._prediction["grand_hamiltonian"]
        self.hamiltonian: float = self._prediction["hamiltonian"]
        self.itfdf_similarity: float = self._prediction["itfdf_similarity"]
        self.matches: list = self._prediction["matches"]
        self.missing: list = self._prediction["missing"]
        self.name: str = self._prediction["name"]
        self.past: list = self._prediction["past"]
        self.potential: float = self._prediction["potential"]
        self.present: list = self._prediction["present"]
        self.similarity: float = self._prediction["similarity"]
        self.snr: float = self._prediction["snr"]
        self.type: str = self._prediction["type"]

        self.events = list(chain(self.past, self.present, self.future))

        if metadata is None:
            metadata = [None for _ in self.events]
        self.metadata = metadata
        pass

    def __repr__(self) -> str:
        return f"Prediction(MODEL|{self.name}, potential={self.potential})"

    def toJSON(self) -> dict:
        """Convert Prediction Object back into dictionary format

        Returns:
            dict: prediction in dict format
        """
        return {
            "confidence": self.confidence,
            "confluence": self.confluence,
            "emotives": self.emotives,
            "entropy": self.entropy,
            "evidence": self.evidence,
            "extras": self.extras,
            "fragmentation": self.fragmentation,
            "frequency": self.frequency,
            "future": self.future,
            "grand_hamiltonian": self.grand_hamiltonian,
            "hamiltonian": self.hamiltonian,
            "itfdf_similarity": self.itfdf_similarity,
            "matches": self.matches,
            "missing": self.missing,
            "name": self.name,
            "past": self.past,
            "potential": self.potential,
            "present": self.present,
            "similarity": self.similarity,
            "snr": self.snr,
            "type": self.type,
        }
        
    def toNumericJSON(self) -> dict:
        """Convert Prediction Object back into dictionary format

        Returns:
            dict: prediction in dict format
        """
        return {
            "confidence": self.confidence,
            "confluence": self.confluence,
            "entropy": self.entropy,
            "evidence": self.evidence,
            "fragmentation": self.fragmentation,
            "frequency": self.frequency,
            "grand_hamiltonian": self.grand_hamiltonian,
            "hamiltonian": self.hamiltonian,
            "itfdf_similarity": self.itfdf_similarity,
            "name": self.name,
            "potential": self.potential,
            "similarity": self.similarity,
            "snr": self.snr,
        }

    def toPastStateGraph(self, starting_idx=0) -> nx.Graph:
        """Make graph from symbols in past of prediction

        Args:
            starting_idx (int, optional): _description_. Defaults to 0.

        Returns:
            nx.Graph: resultant graph
        """
        G = nx.Graph()

        idx = starting_idx
        event: list
        event_indexes = []
        for event in self.past:
            symbol: str
            for symbol in event:
                # node metadata for graph (might not mean much for past state)
                symbol_attributes = {"symbol_type": None}
                if symbol in self.missing:
                    symbol_attributes["symbol_type"] = "missing"
                elif symbol in self.matches:
                    symbol_attributes["symbol_type"] = "matches"
                elif symbol in self.extras:
                    symbol_attributes["symbol_type"] = "extras"
                G.add_node(idx, **symbol_attributes)
                event_indexes.append(idx)
                if idx != 0:
                    G.add_edge(idx - 1, idx)
                idx += 1

        return G

    def toPresentStateGraph(self, starting_idx=0) -> nx.Graph:
        G = nx.Graph()

        idx = starting_idx
        event: list
        event_indexes = []
        for event in self.present:
            symbol: str
            for symbol in event:
                # node metadata for graph
                symbol_attributes = {"symbol_type": None}
                if symbol in self.missing:
                    symbol_attributes["symbol_type"] = "missing"
                elif symbol in self.matches:
                    symbol_attributes["symbol_type"] = "matches"
                elif symbol in self.extras:
                    symbol_attributes["symbol_type"] = "extras"
                G.add_node(idx, **symbol_attributes)
                event_indexes.append(idx)
                if idx != 0:
                    G.add_edge(idx - 1, idx)
                idx += 1

            event_indexes = []
        return G

    def toEventGraph(self) -> nx.Graph:
        G = nx.Graph()

        idx = 0
        past_idxs = []
        present_idxs = []
        future_idxs = []
        for e in self.past:
            past_idxs.append(idx)
            idx += 1
        for e in self.present:
            present_idxs.append(idx)
            idx += 1
        for e in self.future:
            future_idxs.append(idx)
            idx += 1
        print(f"{past_idxs=}, {present_idxs=}, {future_idxs=}")

        G.add_nodes_from(
            zip(past_idxs, [{"data": e, "time": "past"} for e in self.past])
        )
        G.add_nodes_from(
            zip(present_idxs, [{"data": e, "time": "present"}
                for e in self.present])
        )
        G.add_nodes_from(
            zip(future_idxs, [{"data": e, "time": "future"}
                for e in self.future])
        )

        for i, idx in enumerate(past_idxs):
            if i >= len(past_idxs) - 1:
                break
            u, v = past_idxs[i], past_idxs[i + 1]
            attributes = {"time": "past", "weight": 1,
                          "frequency": self.frequency}
            conditional_add_edge(u=u, v=v, graph=G, attributes=attributes)
        for i, event in enumerate(present_idxs):
            if i == 0 and self.past:
                u, v = past_idxs[-1], present_idxs[i]
                attributes = {
                    "time": "past_to_present",
                    "weight": 1,
                    "frequency": self.frequency,
                }
                conditional_add_edge(u=u, v=v, graph=G, attributes=attributes)
            if i >= len(present_idxs) - 1:
                break
            else:
                u, v = present_idxs[i], present_idxs[i + 1]
                attributes = {
                    "time": "present",
                    "weight": 1,
                    "frequency": self.frequency,
                }
                conditional_add_edge(u=u, v=v, graph=G, attributes=attributes)

        for i, event in enumerate(future_idxs):
            if i == 0 and self.present:
                u, v = present_idxs[-1], future_idxs[i]
                attributes = {
                    "time": "present_to_future",
                    "weight": 1,
                    "frequency": self.frequency,
                }
                conditional_add_edge(u=u, v=v, graph=G, attributes=attributes)
            if i == len(self.future) - 1:
                break
            else:
                u, v = future_idxs[i], future_idxs[i + 1]
                attributes = {
                    "time": "future",
                    "weight": 1,
                    "frequency": self.frequency,
                }
                conditional_add_edge(u=u, v=v, graph=G, attributes=attributes)

        return G

    def toLoopingEventGraph(self) -> nx.DiGraph:
        """Produce directed graph from sequence of events.
        Add metadata to edges between events

        Returns:
            nx.DiGraph: resultant graph
        """

        G = nx.DiGraph()

        # add a node for each event, including symbol list, time information in node_data
        event_hashes = [hash_event(e) for e in self.events]
        for event, event_hash in zip(self.events, event_hashes):
            cur_event_details = {"data": event}
            if G.has_node(event_hash):
                continue
            G.add_node(event_hash, **cur_event_details)

        edge_attribute_dict = defaultdict(lambda: defaultdict(list))
        for i, event_hash in enumerate(event_hashes):
            if i == len(event_hashes) - 1:
                break
            u, v = event_hashes[i], event_hashes[i + 1]
            attributes = {
                "metadata": [self.metadata[i]],
            }

            for attr_key, attr_val in attributes.items():
                edge_attribute_dict[u, v][attr_key] += attr_val

        for (u, v), edge_dict in edge_attribute_dict.items():
            G.add_edge(u, v, **edge_dict)

        return G

    def toSymbolGraphs(self) -> Dict[str, nx.DiGraph]:
        symbol_graphs: Dict[str, nx.DiGraph] = {}

        symbol_keys = set()
        event: list
        for event in self.events:
            sym: str
            for sym in event:
                symbol_key = sym.rsplit('|', maxsplit=1)[0]
                symbol_keys.add(symbol_key)

        for symbol_key in symbol_keys:
            symbol_sequence = []
            for event in self.events:
                cur_symbol_values = [sym.rsplit('|', maxsplit=1)[
                    1] for sym in event if f'{symbol_key}|' in sym]
                if not cur_symbol_values:
                    cur_symbol_values = ['null']
                symbol_sequence.append(cur_symbol_values)
            # print(f'Symbol {symbol_key} sequence = {symbol_sequence}')
            symbol_value_counter = Counter(chain(*symbol_sequence))

            symbol_graphs[symbol_key] = nx.DiGraph()
            symbol_graphs[symbol_key].add_nodes_from(
                [(sym, {'count': sym_count}) for sym, sym_count in symbol_value_counter.items()])

            symbol_transition_counter = Counter()
            for i, _ in enumerate(symbol_sequence):
                if i == 0:
                    continue
                for sym in symbol_sequence[i-1]:
                    for new_sym in symbol_sequence[i]:
                        symbol_transition_counter.update([(sym, new_sym)])

            for i, _ in enumerate(symbol_sequence):
                if i == 0:
                    continue

                for sym in symbol_sequence[i-1]:
                    for new_sym in symbol_sequence[i]:
                        symbol_graphs[symbol_key].add_edge(
                            sym, new_sym, weight=symbol_transition_counter[(sym, new_sym)])

        return symbol_graphs


class PredictionEnsemble:
    def __init__(
        self, ensemble, metadata_dict: dict = None, node_name: str = None
    ) -> None:
        """Convert Prediction ensemble into class object

        Args:
            ensemble (_type_): the ensemble in list or dict form
            metadata_dict (dict, optional): dict of {model_name: metadata}. Defaults to None.
            node_name (str, optional): Node name for ensemble if provided as list. Defaults to None.
        """
        if node_name is None:
            node_name = "NODE"

        if isinstance(ensemble, list):
            ensemble = {node_name: ensemble}

        self._ensemble = deepcopy(ensemble)
        self.ensemble = {}

        for k, preds in self._ensemble.items():
            if metadata_dict:
                self.ensemble[k] = [
                    Prediction(prediction=p, metadata=metadata_dict[p["name"]])
                    for p in preds
                ]
            else:
                self.ensemble[k] = [Prediction(prediction=p) for p in preds]

    def __repr__(self) -> str:
        return f"PredictionEnsemble(nodes={list(self.ensemble.keys())})"

    def toDataFrame(self) -> pd.DataFrame:
        """Construct pandas DataFrame using prediction as a row

        Returns:
            pd.DataFrame: showing all predictions from ensemble
        """
        predictions = []
        for k, preds in self.ensemble.items():
            pred: Prediction
            for pred in preds:
                predictions.append({**pred.toJSON(), **{"node": k}})

        return pd.DataFrame(predictions)


    def toNumericDataFrame(self) -> pd.DataFrame:
        """Construct pandas DataFrame using prediction as a row. Use numeric fields and name only

        Returns:
            pd.DataFrame: showing all predictions from ensemble
        """
        predictions = []
        for k, preds in self.ensemble.items():
            pred: Prediction
            for pred in preds:
                predictions.append({**pred.toNumericJSON(), **{"node": k}})

        return pd.DataFrame(predictions)

    def toEventGraph(self) -> nx.DiGraph:
        """Merge Event graphs from each prediction into a single, larger directed graph

        Returns:
            nx.DiGraph: merged directed graph
        """
        pred_graphs = []
        for preds in self.ensemble.values():
            pred: Prediction
            for i, pred in enumerate(preds):
                pred_graphs.append(pred.toLoopingEventGraph())
        graph: nx.DiGraph
        for i, graph in enumerate(pred_graphs):
            for j, (node, node_data) in enumerate(graph.nodes(data=True)):
                node_data["pos"] = (10 * j, i)

        final_graph = nx.DiGraph()

        edge_attribute_dict = defaultdict(lambda: defaultdict(list))
        graph: nx.DiGraph
        for graph in pred_graphs:
            for node, prev_node_data in graph.nodes(data=True):
                # prev_node_data['time'] = None
                if not final_graph.has_node(node):
                    final_graph.add_node(node, **prev_node_data)

            for u, v, prev_edge_data in graph.edges(data=True):
                for attr_key, attr_val in prev_edge_data.items():
                    edge_attribute_dict[u, v][attr_key] += attr_val

        for (u, v), edge_dict in edge_attribute_dict.items():
            final_graph.add_edge(u, v, **edge_dict)

        logger.debug(
            f'Merged Graph: nodes={final_graph.number_of_nodes()}, edges={final_graph.number_of_edges()}')
        return final_graph

    def toSymbolGraphs(self) -> Dict[str, nx.DiGraph]:
        """Merge Symbol graphs from multiple predictions

        Returns:
            Dict[str, nx.DiGraph]: _description_
        """
        pred_graphs = []
        symbol_key_set = set()

        # generate graphs from all predictions
        for preds in self.ensemble.values():
            pred: Prediction
            for i, pred in enumerate(preds):
                pred_graphs.append(pred.toSymbolGraphs())

        # get set of each symbol key available
        for pred_graph_dict in pred_graphs:
            symbol_key_set.update(list(pred_graph_dict.keys()))

        final_graph_dict: Dict[str, nx.DiGraph] = {}

        # iterate across symbol keys, collecting attributes for each node
        for symbol_key in symbol_key_set:
            final_graph_dict[symbol_key] = nx.DiGraph()
            unique_sym_values = set()
            symbol_edge_data = defaultdict(lambda: defaultdict(int))
            for pred_graph_dict in pred_graphs:
                if symbol_key not in pred_graph_dict:
                    continue
                for (u, v), data in (pred_graph_dict[symbol_key]).edges.items():
                    for key, value in data.items():
                        symbol_edge_data[(u, v)][key] += value
                    unique_sym_values.update(pred_graph_dict[symbol_key].nodes)
            logger.debug(f'{symbol_key=}, {unique_sym_values=}')
            final_graph_dict[symbol_key].add_nodes_from(unique_sym_values)

            for (u, v), d in symbol_edge_data.items():
                final_graph_dict[symbol_key].add_edge(u, v, **d)

        return final_graph_dict



import json
import logging

logger = logging.getLogger(__file__)


class Genome:
    """Wrapper class for a Genome topology, loaded from a dict

    :ivar topology: the data representing a Genome
    :ivar agent: agent data parsed from topology
    :ivar description: Genome description field
    :ivar primitive_map: dict of "node name" to "node id" pairings
    """

    def __init__(self, topology: dict):
        """Initialize a genome from a GAIuS agent topology (dict)

        Args:
            topology (dict, required): GAIuS agent topology
        """
        self.topology = topology
        self.agent = self.topology['agent']
        self.description = self.topology['description']
        # Not everyone will have style information, so wrap this:
        try:
            self.style_obj = self.topology['style']
            self.style_obj[1]['style']['curve-style'] = 'bezier'
        except Exception:
            pass
        self.primitives = {}
        self.manipulatives = {}
        self.action_ids = []
        self.actions_manifests = {}
        for node in self.topology['elements']['nodes']:
            if node['data']['type'] == 'primitive':
                self.primitives[node['data']['id']] = node['data']
            elif node['data']['type'] == 'manipulative':
                self.manipulatives[node['data']['id']] = node['data']

        self.agent_genome = {'primitives': self.primitives,
                             'manipulatives': self.manipulatives}
        self.primitive_map = {x['name']: _id for _id,
                              x in self.primitives.items()}
        self.manipulative_map = {_id: x['name']
                                 for _id, x in self.manipulatives.items()}

        return

    def get_nodes(self):
        """Return a tuple of :samp:`primitive node names`, :samp:`manipulative node names` from the topology"""
        return self.agent_genome['primitives'], self.agent_genome['manipulatives']

    def get_primitive_map(self):
        """Get a map of node names to primitive ids in the GAIuS agent topology

        Example:
            .. code-block:: python

                >>> agent = AgentClient(agent_info)
                >>> agent.connect()
                >>> agent.genome.get_primitive_map()
                {'P1': 'p46b6b076c'}
        """
        return self.primitive_map

    def get_manipulative_map(self):
        """Get a map of the manipulatives connected to each primitive node in the GAIuS agent topology

        Returns:
            dict: dictionary of manipulatives mappings in the topology
        """
        return self.manipulative_map

    def change_genes(self, p_id, gene_data):
        """Change the genes of a primitive in the Genome's cache

        Args:
            p_id (str): primitive id of the node to edit
            gene_data (dict): dictionary of genes to update in the Genome's cache
        """
        for key, value in gene_data.items():
            self.agent_genome['primitives'][p_id][key] = value
        return

    def display(self):  # pragma: no cover
        """Display the Genome topology in Cytoscape
        """
        try:
            from ipycytoscape import CytoscapeWidget
            from IPython.display import display, clear_output
            from ipywidgets import Output
        except ImportError as e:
            logger.exception(
                "Failed to import dependencies for displaying Cytoscape graph")
            raise

        cyto = CytoscapeWidget()
        cyto.graph.add_graph_from_json(self.topology['elements'])
        cyto.set_style(self.style_obj)
        clear_output()

        out = Output(layout={'border': '1px solid black'})

        def display_node_data(node):
            """Show node data on click
            """
            with out:
                out.clear_output()
                print(
                    f'Node {node["data"]["id"]} details:\n{json.dumps(node, indent=2)}')
            pass

        def display_edge_data(edge):
            """Show edge data on click
            """
            with out:
                out.clear_output()
                print(
                    f'Edge {edge["data"]["id"]} details:\n{json.dumps(edge, indent=2)}')
            pass

        cyto.on('node', 'click', display_node_data)
        cyto.on('edge', 'click', display_edge_data)
        display(cyto)
        display(out)

        return



#!/usr/bin/env python3
import os
import docker
import docker.errors as de
import json
import logging
import platformdirs as plat
import shutil
import uuid
import subprocess
import logging
from pathlib import Path
from typing import List, Dict, Union
from ia.gaius.agent_client import AgentClient
from ia.gaius.experimental.comcom_client import COMCOMClient
from ia.scripts.spawn_agent import stopit, start_agent, delete_docker_network
from ia.scripts.spawn_general import spawn_container, kill_container, \
    connect_networks_container, get_docker_bridge_ip, create_docker_network
from contextlib import contextmanager

logger = logging.getLogger(__name__)


class AgentInfo:
    def __init__(self,
                 user_id: str,
                 agent_id: str,
                 agent_name: str,
                 genome_name: str = None,
                 genome_file: str = None,
                 agent_config=None,
                 location: str = 'local',
                 LOG_LEVEL=logging.INFO,
                 api_key="ABCD-1234",
                 genome_dir=None,
                 agents_dir=None,
                 genome=None,
                 single_container: bool = False,
                 local_agent_config=None,
                 **kwargs):
        """Wrapper class for Agent information. Used to start/stop agents, and save config information to a central location

        Args:
            user_id (str): User id attached to docker contaienrs spawned using AgentInfo
            agent_id (str): Agent id attached to docker containers spawned using AgentInfo
            agent_name (str): Descriptive, user-friendly name to refer to agent by
            genome_name (str, optional): If genome already in centralized location, refer to it by its name
            genome_file (str, optional): To copy outside genome into genome store, provide filepath here
            agent_config (_type_, optional): Stores AgentConfig details once agent is spawned.
            location (str, optional): In future will support spawning at several locations (local, private cloud, DigitalOcean, etc). Defaults to 'local'.
            LOG_LEVEL (logging.LEVEL, optional): Used to control verbosity of messages in agent spawn script. Defaults to logging.INFO.
            api_key (str, optional): Used to set API KEY of spawned agent. Defaults to "ABCD-1234"
        """
        self.data_dir = plat.user_data_dir(
            appname="IA_SDK_AgentManager", appauthor="IntelligentArtifacts")
        self.genome_dir = f'{self.data_dir}/genomes'
        if genome_dir:
            self.genome_dir = genome_dir
        self.agents_dir = f'{self.data_dir}/agents'
        if agents_dir:
            self.agents_dir = agents_dir

        if genome_name:
            self.genome_name = genome_name
        else:
            self.genome_name = os.path.basename(genome_file)
            if os.path.exists(f'{self.genome_dir}/{self.genome_name}'):
                logger.debug(
                    f'copying genome to genome dir would overwrite current genome {os.path.basename(genome_file)}, using genome already present')
            else:
                shutil.copyfile(src=genome_file,
                                dst=f'{self.genome_dir}/{self.genome_name}')

        self.genome_file = f'{self.genome_dir}/{self.genome_name}'
        self.user_id = user_id
        self.agent_id = agent_id
        self.location = location
        self.agent_name = agent_name
        self.agent_config = agent_config
        self.local_agent_config = local_agent_config
        self.api_key = api_key
        self.single_container = single_container

        self.genome = genome
        if self.genome is None:
            with open(self.genome_file) as f:
                self.genome = json.load(f)
                if self.single_container:
                    self.genome['communication_protocol'] = 'uds'

    def check_running(self):
        """Check if we can access the agent

        Returns:
            bool: True for success, False for failure
        """
        if self.location == 'local':
            agent = AgentClient(bottle_info=self.agent_config)
            try:
                agent.set_timeout(10)
                agent.connect()
            except Exception:
                return False

            return True

        else:
            raise Exception(f'unknown location {self.location}')

    def delete_config_file(self):
        """Kill agent and delete config information from agents directory
        """
        self.kill_agent()
        if os.path.exists(f'{self.agents_dir}/{self.agent_name}'):
            os.remove(f'{self.agents_dir}/{self.agent_name}')
        return f"deleted {self.agents_dir}/{self.agent_name}"

    def save_config_file(self):
        """Store config information for agent in agents directory, based on agent name
        """
        with open(f'{self.agents_dir}/{self.agent_name}', 'w') as f:
            json.dump(obj=self.toJSON(), fp=f)

        logger.info(f'saved to {self.agents_dir}/{self.agent_name}')
        return f'{self.agents_dir}/{self.agent_name}'

    def kill_agent(self):
        """Attempt to kill a running agent
        """
        stopit(self.genome, user_id=self.user_id, agent_id=self.agent_id)

    def spawn(self, connect_jia=True):
        """Start a new agent using information provided in constructor

        Args:
            api_key (str, optional): API Key for agent. Defaults to "ABCD-1234".
        """
        # with open(self.genome_file, 'r') as f:
        #     genome = json.load(f)

        json_obj = {'agent_id': self.agent_id,
                    'user_id': self.user_id,
                    'kill': False,
                    'network': None,
                    'api_key': self.api_key,
                    'genome': self.genome}

        self.save_config_file()
        try:
            self.agent_config, ports = start_agent(json_obj=json_obj)
            self.local_agent_config = {'name': '',
                                       'domain': f'localhost:{ports["80/tcp"]}',
                                       'api_key': self.api_key,
                                       'secure': False}
            logger.debug(
                f'updating config file with new agent config: {self.agent_config}')
            self.save_config_file()
        except de.DockerException as spawn_error:
            logger.exception(
                "Received exception while spawning agent. Cleaning up and re-raising")
            self.delete_config_file()
            raise spawn_error

        if connect_jia:
            self.connect_jia()

        return json.loads(json.dumps(self.agent_config))

    def get_docker_networks(self):
        """Get names of networks this agent is a part of (uses gaius-api-{user_id}-{agent_id})

        Raises:
            Exception: More that one container found with same name

        Returns:
            _type_: list of networks
        """
        docker_client = docker.from_env()
        gapi_container = docker_client.containers.list(
            filters={'name': f'gaius-api-{self.user_id}-{self.agent_id}'})
        gapi_container = [container for container in gapi_container if container.name ==
                          f'gaius-api-{self.user_id}-{self.agent_id}']
        if len(gapi_container) == 1:
            gapi_container = gapi_container[0]
            return list(gapi_container.attrs['NetworkSettings']['Networks'].keys())
        elif len(gapi_container) > 1:
            self.delete_config_file()
            raise Exception(
                'too many docker containers with same GAIuS API name')
        else:
            raise Exception('gaius-api container not found, start agent first')

    def connect_jia(self):
        """Connect agent to jia (add jia to the agent's network)
        """
        docker_client = docker.from_env()
        jias = docker_client.containers.list()
        jias = [container for container in jias if any(
            ['registry.digitalocean.com/intelligent-artifacts/jia' in tag for tag in container.image.tags])]
        nets = self.get_docker_networks()
        nets = docker_client.networks.list(names=nets)
        if len(jias) == 0:
            logger.info("No JIA notebooks found! Skipping...")
        elif len(jias) == 1:
            logger.info(f'connecting single jia')
            jia = docker_client.containers.get(
                jias[0].short_id)  # .attrs['Config']['Name']
            for net in nets:
                net.reload()
                if jia.id in [cont.id for cont in net.containers]:
                    logger.info(f'jia already connected to network {net.name}')
                    continue
                net.connect(jia)
                logger.info(f"Connecting JIA id: {jia.name} on {net.name}")
        elif len(jias) > 1:
            def connect_jia_func():
                logger.info(
                    "More than one JIA found. Please select which to connect:")
                for c, i in enumerate(jias):
                    x = int(input(f"    {i}: {c} {c.name} connect? Y/N - "))
                    if x.lowercase() == 'y':
                        for net in nets:
                            net.reload()
                            if c.id in [cont.id for cont in net.containers]:
                                logger.info(
                                    f'jia already connected to network {net.name}')
                                continue
                            net.connect(c)
                            logger.info(
                                f"Connecting JIA id: {c.name} on {net.name}")
                    else:
                        logger.info("Invalid entry. Try again...")

    def get_agent_client(self, local: bool = False):
        """Retreive AgentClient object from started agent

        Returns:
            ia.gaius.agent_client.AgentClient
        """
        if self.agent_config == None:
            raise Exception(
                'agent_config information empty, must start agent first')
        elif local:
            return AgentClient(bottle_info=self.local_agent_config)

        return AgentClient(bottle_info=self.agent_config)

    def toJSON(self):
        """Dump config information to a JSON object (dict)

        Returns:
            dict: config information
        """
        return {'genome_name': self.genome_name,
                'genome_file': self.genome_file,
                'agent_id': self.agent_id,
                'user_id': self.user_id,
                'agent_name': self.agent_name,
                'location': self.location,
                'agent_config': self.agent_config,
                'local_agent_config': self.local_agent_config,
                'api_key': self.api_key,
                'genome_dir': str(self.genome_dir),
                'agents_dir': str(self.agents_dir),
                'genome': json.dumps(self.genome),
                'single_container': self.single_container
                }

    @classmethod
    def fromFile(cls, filepath: Union[str, Path]):
        """Construct AgentInfo object from file

        Args:
            filepath (str): path to file to load

        Returns:
            AgentInfo: AgentInfo object created from the file
        """
        with open(filepath, 'r') as f:
            json_obj = json.load(f)

        return cls.fromJSON(json_obj=json_obj)

    @classmethod
    def fromJSON(cls, json_obj):
        """Construct AgentInfo from JSON object

        Args:
            json_obj (dict): Contains fields necessary for constructing AgentInfo

        Returns:
            AgentInfo: AgentInfo object created from JSON object
        """
        return cls(**json_obj)


class TFInfo:
    def __init__(self,
                 user_id: str,
                 tf_id: str,
                 tf_name: str,
                 location: str,
                 api_key: str,
                 thinkflux_dir: str,
                 docker_image: str,
                 port: int,
                 privileged: bool = False,
                 environment: dict = None,
                 volumes: dict = None,
                 **kwargs) -> None:

        if volumes is None:
            volumes = {}
        if environment is None:
            environment = {}

        self.user_id = user_id
        self.tf_id = tf_id
        self.tf_name = tf_name
        self.location = location
        self.api_key = api_key
        self.thinkflux_dir = thinkflux_dir
        self.docker_image = docker_image
        self.port = port
        self.privileged = privileged
        self.volumes = volumes
        self.environment = environment
        self.container_name = f'tf-{user_id}-{tf_id}'

    def spawn(self) -> bool:

        if spawn_container(container_name=self.container_name,
                           api_key=self.api_key,
                           ports={'8080/tcp': self.port},
                           docker_image=self.docker_image,
                           env=self.environment,
                           volumes=self.volumes):
            self.save_config_file()
            return True

        return False

    def kill(self):
        kill_container(self.container_name)
        return self.delete_config_file()

    def delete_config_file(self):
        """delete config information from agents directory
        """
        if os.path.exists(f'{self.thinkflux_dir}/{self.tf_name}'):
            os.remove(f'{self.thinkflux_dir}/{self.tf_name}')
        return f"deleted {self.thinkflux_dir}/{self.tf_name}"

    def save_config_file(self):
        """Store config information for agent in agents directory, based on agent name
        """
        with open(f'{self.thinkflux_dir}/{self.tf_name}', 'w') as f:
            json.dump(obj=self.toJSON(), fp=f)

        logger.debug(f'saved to {self.thinkflux_dir}/{self.tf_name}')
        return f'{self.thinkflux_dir}/{self.tf_name}'

    def connect_agents(self, agents: List[AgentInfo]):
        nets = set()
        for agent in agents:
            nets.update(agent.get_docker_networks())

        logger.debug(f'{nets=}')
        if not connect_networks_container(networks=list(nets), container_name=self.container_name):
            return False

        return True

    def toJSON(self):
        """Dump config information to a JSON Object

        Returns:
            dict: config information
        """
        return {'user_id': self.user_id,
                'tf_id': self.tf_id,
                'tf_name': self.tf_name,
                'location': self.location,
                'api_key': self.api_key,
                'port': self.port,
                'thinkflux_dir': str(self.thinkflux_dir),
                'docker_image': self.docker_image,
                'environment': self.environment,
                'privileged': self.privileged,
                'volumes': self.volumes
                }

    @classmethod
    def fromFile(cls, filepath: Union[str, Path]):
        """Construct AgentInfo object from file

        Args:
            filepath (str): path to file to load

        Returns:
            AgentInfo: AgentInfo object created from the file
        """
        with open(filepath, 'r') as f:
            json_obj = json.load(f)

        return cls.fromJSON(json_obj=json_obj)

    @classmethod
    def fromJSON(cls, json_obj):
        """Construct AgentInfo from JSON object

        Args:
            json_obj (dict): Contains fields necessary for constructing AgentInfo

        Returns:
            AgentInfo: AgentInfo object created from JSON object
        """
        return cls(**json_obj)


class ComcomInfo:
    def __init__(self,
                 comcom_id: str,
                 name: str,
                 location: str,
                 api_key: str,
                 save_dir: str,
                 docker_image: str,
                 port: int,
                 ) -> None:
        """Construct a ComcomInfo object. Used to manage spawning and connecting
        Comcom docker containeres with agents and thinkflux instances

        Args:
            comcom_id (str): Unique identifier attached to Comcom docker container spawned using this object
            name (str): Human friendly name used to retrieve ComcomInfo from AgentManager
            location (str): Where to spawn this Comcom. Currently only supports "local"
            api_key (str): api key passed to docker container in environment variable
            save_dir (str): where ComcomInfo objects are to be saved (as file)
            docker_image (str): Specific docker image to utilize when spawning Comcom
            port (int): Port which Comcom will be connected to
        """

        self.comcom_id = comcom_id
        self.name = name
        self.location = location
        self.api_key = api_key
        self.port = port
        self.save_dir = save_dir
        self.docker_image = docker_image
        self.container_name = f'Comcom-{comcom_id}'
        self._mongo_container_name = f'{self.container_name}-kb'
        self._network_name=f'g2network-{self.container_name}'

    def spawn(self) -> bool:

        ports = list(range(7510, 7811))
        port_udps = [f'{port}/udp' for port in ports]
        port_range_dict = dict(zip(port_udps, ports))
        
        # docker_bridge_ip = get_docker_bridge_ip()
        comcom_ports = {'8081/tcp': self.port,
                        }
        comcom_ports.update(port_range_dict)
        
        try:
            spawn_container(container_name=self.container_name,
                        api_key=self.api_key,
                        ports=comcom_ports,
                        docker_image=self.docker_image)
        
            logger.debug("COMCOM spawned, now spawning a MongoDB container")
        
            spawn_container(container_name=self._mongo_container_name,
                        hostname='comcom-kb',
                        api_key='b1',
                        docker_image='mongo:latest',
                        ports={})
        
            create_docker_network(network_name=self._network_name)
            
            connect_networks_container(networks=[self._network_name], container_name=self.container_name)
            connect_networks_container(networks=[self._network_name], container_name=self._mongo_container_name)
        except Exception as e:
            logger.exception("Failed to start Comcom")
            self.kill()
            raise

        self.save_config_file()

        return True

    def get_comcom_client(self, local: bool = False) -> COMCOMClient:
        """Create A COMCOMClient instance corresponding to the spawned Comcom 

        Args:
            local (bool, optional): Whether to use container_name or localhost to connect. Defaults to False.

        Returns:
            COMCOMClient
        """
        domain = f"{self.container_name}:{self.port}"
        if local:
            domain = f"localhost:{self.port}"
            
        comcom_details = {
            "name" : "",
            "domain" : domain,
            "api_key" : self.api_key,
            "secure" : True
        }
        return COMCOMClient(comcom_details, verify=False)
        
    def kill(self):
        """Kill Comcom and delete container
        """
        kill_container(self.container_name)
        kill_container(self._mongo_container_name)
        delete_docker_network(docker_client=docker.from_env(), network_name=self._network_name)
        return self.delete_config_file()

    def delete_config_file(self):
        """delete config file information
        """
        if os.path.exists(f'{self.save_dir}/{self.name}'):
            os.remove(f'{self.save_dir}/{self.name}')
        return f"deleted {self.save_dir}/{self.name}"

    def save_config_file(self):
        """Store config information, based on comcom name
        """
        with open(f'{self.save_dir}/{self.name}', 'w') as f:
            json.dump(obj=self.toJSON(), fp=f)

        logger.debug(f'saved to {self.save_dir}/{self.name}')
        return f'{self.save_dir}/{self.name}'

    def connect_agents(self, agents: List[AgentInfo]):
        """Connect agents to Comcom docker network

        Args:
            agents (List[AgentInfo]): list of AgentInfo objects to connect to network
        """
        nets = set()
        for agent in agents:
            nets.update(agent.get_docker_networks())

        logger.debug(f'{nets=}')
        if not connect_networks_container(networks=list(nets), container_name=self.container_name):
            return False

        return True

    def toJSON(self):
        """Dump config information to a JSON Object

        Returns:
            dict: config information
        """
        return {
            'comcom_id': self.comcom_id,
            'name': self.name,
            'location': self.location,
            'api_key': self.api_key,
            'port': self.port,
            'save_dir': str(self.save_dir),
            'docker_image': self.docker_image,
        }

    @classmethod
    def fromFile(cls, filepath: Union[str, Path]):
        """Construct AgentInfo object from file

        Args:
            filepath (str): path to file to load

        Returns:
            AgentInfo: AgentInfo object created from the file
        """
        with open(filepath, 'r') as f:
            json_obj = json.load(f)

        return cls.fromJSON(json_obj=json_obj)

    @classmethod
    def fromJSON(cls, json_obj):
        """Construct AgentInfo from JSON object

        Args:
            json_obj (dict): Contains fields necessary for constructing AgentInfo

        Returns:
            AgentInfo: AgentInfo object created from JSON object
        """
        return cls(**json_obj)


class AgentManager:
    def __init__(self,
                 genome_dir: Path = None,
                 agents_dir: Path = None,
                 thinkflux_dir: Path = None,
                 comcom_dir: Path = None,
                 local: bool = False):
        """Initialize AgentManager object (setup dirs, etc.)

        This object can be used to spawn Agents using the start_agent() method.
        """
        self.data_dir = plat.user_data_dir(
            appname="IA_SDK_AgentManager", appauthor="IntelligentArtifacts")
        self.genome_dir = Path(f'{self.data_dir}/genomes')
        if genome_dir:
            self.genome_dir = Path(genome_dir)
        self.agents_dir = Path(f'{self.data_dir}/agents')
        if agents_dir:
            self.agents_dir = Path(agents_dir)

        self.thinkflux_dir = Path(f'{self.data_dir}/thinkflux')
        if thinkflux_dir:
            self.thinkflux_dir = Path(thinkflux_dir)

        self.comcom_dir = Path(f'{self.data_dir}/comcom')
        if comcom_dir:
            self.comcom_dir = Path(comcom_dir)

        self.config_file = f'{self.data_dir}/config.json'
        self.local = local
        if not os.path.exists(self.data_dir):
            os.makedirs(self.data_dir, exist_ok=True)
        logger.info(f'   agents dir: {self.agents_dir}')
        logger.info(f'   genome dir: {self.genome_dir}')
        logger.info(f'thinkflux dir: {self.thinkflux_dir}')
        logger.info(f'   comcom_dir: {self.comcom_dir}')

        if not os.path.exists(self.genome_dir):
            os.makedirs(self.genome_dir, exist_ok=True)

        if not os.path.exists(self.agents_dir):
            os.makedirs(self.agents_dir, exist_ok=True)

        if not os.path.exists(self.thinkflux_dir):
            os.makedirs(self.thinkflux_dir, exist_ok=True)

        if not os.path.exists(self.comcom_dir):
            os.makedirs(self.comcom_dir, exist_ok=True)

        self.current_agents: Dict[str, AgentInfo] = {}
        self.current_tfs: Dict[str, TFInfo] = {}
        self.current_comcoms: Dict[str, ComcomInfo] = {}

        self.update_current_agents()

    def start_hoster(self):
        """Start `dvdarias/docker-hoster <https://github.com/dvddarias/docker-hoster>`_ docker container to resolve
        container hostnames into IP addresses. More friendly for AgentManager usage"""
        START_COMMAND = """
        if [ ! "$(docker ps -a -q -f name=docker-hoster)" ]; then
            if [ "$(docker ps -aq -f status=exited -f name=docker-hoster)" ]; then
                # cleanup
                docker rm docker-hoster
            fi
            # run your container
            
            docker run -d --rm\
                    -v /var/run/docker.sock:/tmp/docker.sock \
                    -v /etc/hosts:/tmp/hosts \
                    --name docker-hoster \
                    dvdarias/docker-hoster
        fi
        """
        output = subprocess.run(START_COMMAND, capture_output=True, shell=True)
        if output.returncode != 0:
            raise Exception(
                f'Starting hoster failed with exitcode {output.returncode}')
        return

    def stop_hoster(self):
        """Stop `dvdarias/docker-hoster <https://github.com/dvddarias/docker-hoster>`_ docker container"""
        STOP_COMMAND = """
        if [ "$(docker ps -a -q -f name=docker-hoster)" ]; then
            docker rm --force docker-hoster;
        fi
        
        """
        output = subprocess.run(STOP_COMMAND,
                                capture_output=True,
                                shell=True)
        if output.returncode != 0:
            raise Exception(
                f'Stopping hoster failed with exitcode {output.returncode}')
        return

    def start_tf(self,
                 tf_name: str,
                 api_key: str,
                 docker_image: str,
                 port: int = 8090,
                 user_id: str = 'tf',
                 tf_id: str = '1',
                 environment: dict = None,
                 privileged: bool = None,
                 volumes: dict = None,
                 agents_to_connect: list = None) -> TFInfo:
        """Spawn a Thinkflux container using provided config. Store config in AgentManager

        Container name is defined as:
            "tf-{user_id}-{tf_id}"

        Args:
            tf_name (str): descriptive name "alias" for Thinkflux
            api_key (str): Passed as API_KEY env variable to container
            docker_image (str): Container image to spawn. E.g. "registry.digitalocean.com/intelligent-artifacts/thinkflux:develop"
            port (int, optional): _description_. Defaults to 8090.
            user_id (str, optional): user_id portion of container extension. Defaults to 'tf'.
            tf_id (str, optional): tf_id portion of container extension. Defaults to '1'.
            environment (dict, optional): additional environment variables to provide to container, as key-value pairs. Defaults to None.
            privileged (bool, optional): run container in privileged mode. Defaults to None.
            volumes (dict, optional): volumes to mount in container as key-value pairs. Defaults to None.
            agents_to_connect (list, optional): list of agents to connect to Thinkflux. Defaults to None.

        Raises:
            Exception: If invalid agents passed in agents_to_connect
            Exception: If tf_name already exists in AgentManager
            Exception: If a thinkflux with user_id and tf_id already exists in AgentManager

        Returns:
            TFInfo: Object corresponding to spawned Thinkflux
        """

        self.update_current_agents()

        agents = []

        if agents_to_connect is None:
            agents_to_connect = []

        for agent_name in agents_to_connect:
            if isinstance(agent_name, AgentInfo):
                agents.append(agent_name)

            elif isinstance(agent_name, str):

                if agent_name not in self.current_agents:
                    raise KeyError(
                        f'Agent {agent_name} not found in current agents')
                agents.append(self.current_agents[agent_name])

            else:
                raise Exception(
                    f"Invalid agent found in agents_to_connect: {agent_name}")

        if tf_name in self.current_tfs:
            raise Exception(f'Thinkflux({tf_name}) already exists')
            pass

        for name, tf in self.current_tfs.items():
            if tf.tf_id == tf_id and tf.user_id == user_id:
                raise Exception(
                    f'Thinkflux({name}) with user_id {user_id} and tf_id {tf_id} already exists, please choose other ids')

        tf = TFInfo(user_id=user_id,
                    tf_id=tf_id,
                    tf_name=tf_name,
                    location='local',
                    api_key=api_key,
                    thinkflux_dir=self.thinkflux_dir,
                    docker_image=docker_image,
                    port=port,
                    environment=environment,
                    privileged=privileged,
                    volumes=volumes)

        tf.spawn()
        self.update_current_agents()  # update agents after spawn, before linking networks
        logger.debug(f'{agents=}')
        tf.connect_agents(agents=agents)

        return tf

    def delete_tf(self,
                  tf_name: str):
        """Kill a Thinkflux container using it's tf_name attribute

        Args:
            tf_name (str): attribute passed when spawning Thinkflux

        Raises:
            Exception: When config not found
        """
        if tf_name not in self.current_tfs:
            raise Exception(f'Thinkflux({tf_name}) not found in current_tfs')

        tf: TFInfo = self.current_tfs[tf_name]
        tf.kill()

        self.update_current_agents()

    def start_comcom(self,
                     name: str,
                     api_key: str,
                     docker_image: str,
                     port: int = 8081,
                     comcom_id: str = '1',
                     agents_to_connect: list = None) -> ComcomInfo:

        self.update_current_agents()

        agents = []

        if agents_to_connect is None:
            agents_to_connect = []

        for agent_name in agents_to_connect:
            if isinstance(agent_name, AgentInfo):
                agents.append(agent_name)

            elif isinstance(agent_name, str):
                if agent_name not in self.current_agents:
                    raise KeyError(
                        f'Agent {agent_name} not found in current agents')
                agents.append(self.current_agents[agent_name])

        if name in self.current_comcoms:
            raise Exception(f'Comcom({name}) already exists')
            pass

        for name, comcom in self.current_comcoms.items():
            if comcom.comcom_id == comcom_id:
                raise Exception(
                    f'Comcom({name}) with comcom_id {comcom_id} already exists, please choose other ids')

        comcom = ComcomInfo(comcom_id=comcom_id,
                            name=name,
                            location='local',
                            api_key=api_key,
                            save_dir=self.comcom_dir,
                            docker_image=docker_image,
                            port=port
                            )

        comcom.spawn()
        self.update_current_agents()  # update agents after spawn, before linking networks
        logger.debug(f'{agents=}')
        try:
            comcom.connect_agents(agents=agents)
        except Exception as connect_agents_error:
            logger.exception(f'Failed to connect agents to comcom')
            comcom.kill()
            raise

        return comcom

    def delete_comcom(self,
                      name: str):
        if name not in self.current_comcoms:
            raise Exception(
                f'Comcom({name}) not found in current_comcoms')

        comcom: ComcomInfo = self.current_comcoms[name]
        comcom.kill()

        self.update_current_agents()

    def start_agent(self,
                    genome_file=None,
                    genome_name=None,
                    user_id: str = 'jia',
                    agent_id: str = '1',
                    agent_name=None,
                    connect_jia=True,
                    api_key='ABCD-1234',
                    single_container=False) -> AgentInfo:
        """Spawn an agent and save its information in the localized agents dir

        .. note:: 

            If you wish to use a genome that has not been used by the AgentManager before,
            pass it in the genome_file field. This will retrieve the genome from the specified path,
            and save it in the localized genome dir for future use.


        Args:
            genome_file (str, optional): Path to outside genome file. Defaults to None.
            genome_name (str, optional): Name of genome in the localized genome dir. Defaults to None.
            user_id (str, optional): user id for naming docker containers. Defaults to 'jia'.
            agent_id (str, optional): agent id for naming docker containers. Defaults to '1'.
            agent_name (_type_, optional): descriptive name "alias" for agent. Defaults to None.
            connect_jia (bool, optional): Whether to attempt to connect agent to Jia docker container. Defaults to True.
            api_key (str, optional): Control the api key used to access agent. Defaults to 'ABCD-1234'.

        Raises:
            Exception: If there is a conflicting agent with user_id and agent_id already present/alive

        Returns:
            AgentInfo: AgentInfo object corresponding to started agent. can be used to retrieve
            AgentClient object using AgentInfo.get_agent_client()
        """
        if agent_name is None:
            agent_name = f'agent-{uuid.uuid4().hex[:8]}'
        self.update_current_agents()
        for name, agent in self.current_agents.items():
            if agent.agent_id == agent_id and agent.user_id == user_id:
                raise Exception(
                    f'Agent({name}) with user_id {user_id} and agent_id {agent_id} already exists, please choose other ids')

        if agent_name is None:
            f'agent-{uuid.uuid4().hex[:8]}'
        agent = AgentInfo(agent_id=agent_id,
                          user_id=user_id,
                          genome_file=genome_file,
                          genome_name=genome_name,
                          agent_name=agent_name,
                          api_key=api_key,
                          genome_dir=str(self.genome_dir),
                          agents_dir=str(self.agents_dir),
                          single_container=single_container)
        try:
            agent.spawn(connect_jia=connect_jia)
        except Exception:
            logger.exception("Failed to spawn agent, cleaning up")
            agent.delete_config_file()
            raise
        self.update_current_agents()
        logger.info(f'started agent {agent.agent_name}')
        return agent

    def get_all_agent_status(self):
        """Get the status of all agents in current_agents dict

        Returns:
            dict: { agent_name : alive? }
        """
        agent_status = {}
        for agent_name, agent in self.current_agents.items():
            try:
                agent_status[agent_name] = agent.check_running()
            except Exception as e:
                logger.exception(
                    f'Error checking status of agent {agent_name}: {str(e)}')
                agent_status[agent_name] = False
                pass

        return agent_status

    def delete_agent(self, agent_name):
        """Stop/Delete agent & remove config file from agents_dir

        Args:
            agent_name (str): name of agent to stop

        Returns:
            str: string depicting deletion status
        """
        result = self.current_agents[agent_name].delete_config_file()
        del self.current_agents[agent_name]
        self.update_current_agents()
        return result

    def list_genomes(self):
        """List the genomes present in the localized genome store

        Returns:
            list: list of genome files 
        """
        return os.listdir(self.genome_dir)

    def delete_genome(self, genome_name):
        """Delete genome from centralized genome_dir

        Args:
            genome_name (str): genome name to delete
        """
        for name, agent in self.current_agents.items():
            if agent.genome_name == genome_name:
                raise Exception(
                    f'Genome {genome_name} in use by Agent {name}. Please kill agent before removing genome')
        genome_path = os.path.join(self.genome_dir, genome_name)
        logger.debug(f'will remove genome {genome_path}')
        os.remove(genome_path)
        return

    def get_genome(self, genome_name):
        """Get genome by name, return as dict

        Args:
            genome_name (str): filename for genome in genome_dir

        Raises:
            Exception: When genome not found

        Returns:
            dict: contents of genome file
        """
        genome_path = os.path.join(self.genome_dir, genome_name)

        if genome_name not in self.list_genomes():
            raise Exception(
                f'{genome_name} not in genome dir: {self.list_genomes()}')
        with open(genome_path, 'r') as f:
            return json.load(f)

    def update_current_agents(self):
        """Update AgentManager's list of configured agents, thinkflux, comcom, genomes
        """

        # retrieve all agent files
        self.current_agents = os.listdir(self.agents_dir)
        self.current_tfs = os.listdir(self.thinkflux_dir)
        self.current_comcoms = os.listdir(self.comcom_dir)

        # get full paths of agent files
        self.current_agents = [self.agents_dir.joinpath(
            agent) for agent in self.current_agents]

        self.current_tfs = [self.thinkflux_dir.joinpath(
            tf) for tf in self.current_tfs]

        self.current_comcoms = [self.comcom_dir.joinpath(
            comcom) for comcom in self.current_comcoms]

        # instantiate AgentInfo classes
        self.current_agents = [AgentInfo.fromFile(
            agent) for agent in self.current_agents]

        self.current_tfs = [TFInfo.fromFile(
            tf) for tf in self.current_tfs]

        self.current_comcoms = [ComcomInfo.fromFile(
            comcom) for comcom in self.current_comcoms]

        # make current agents into a dict by agent_name
        self.current_agents = {
            agent.agent_name: agent for agent in self.current_agents}
        self.current_tfs = {tf.tf_name: tf for tf in self.current_tfs}
        self.current_comcoms = {
            comcom.name: comcom for comcom in self.current_comcoms}

    def remediate_dead_agents(self):
        """Remove all agents that cannot be pinged.
        """

        agent_status = self.get_all_agent_status()

        killed_agents = []
        for agent_name, alive in agent_status.items():
            if not alive:
                logger.debug(f'{agent_name} is not alive, cleaning up')
                self.delete_agent(agent_name)
                killed_agents.append(agent_name)

        logger.debug(f'{killed_agents=}')

    def kill_all_agents(self, update=True):
        """Kill all agents known to AgentManager

        Args:
            update (bool, optional): Whether to update AgentManager information before killing containers. Defaults to True.
        """

        if update:
            self.update_current_agents()

        # We modify the dicts during iteration,
        # so we use this funky syntax
        for agent_name in list(self.current_agents.keys()):
            logger.info(f"Killing agent {agent_name}")
            self.delete_agent(agent_name=agent_name)

        for tf_name in list(self.current_tfs.keys()):
            self.delete_tf(tf_name=tf_name)

        for comcom_name in list(self.current_comcoms.keys()):
            self.delete_comcom(name=comcom_name)

    @contextmanager
    def agent_context(self,
                      genome_file: str = None,
                      genome_name: str = None,
                      user_id: str = 'jia',
                      agent_id: str = '1',
                      agent_name: str = None,
                      connect_jia=True,
                      api_key='ABCD-1234'):
        """Provide ability to use Agent as a context manager. Internally calls start_agent and stop_agent.

        Args:
            genome_file (_type_, optional): _description_. Defaults to None.
            genome_name (_type_, optional): _description_. Defaults to None.
            user_id (str, optional): _description_. Defaults to 'jia'.
            agent_id (str, optional): _description_. Defaults to '1'.
            agent_name (_type_, optional): _description_. Defaults to None.
            connect_jia (bool, optional): _description_. Defaults to True.
            api_key (str, optional): _description_. Defaults to 'ABCD-1234'.

        Yields:
            _type_: _description_
        """
        try:
            agent = self.start_agent(genome_file=genome_file,
                                     genome_name=genome_name,
                                     user_id=user_id,
                                     agent_id=agent_id,
                                     agent_name=agent_name,
                                     connect_jia=connect_jia,
                                     api_key=api_key)
            yield agent.get_agent_client(local=self.local)
        finally:
            self.delete_agent(agent.agent_name)



import functools
import json
import operator
from typing import Union
from copy import deepcopy
from collections import Counter, defaultdict
from itertools import chain
import os
from math import isnan
from pathlib import Path
from tqdm.auto import tqdm
import logging
# Visualization
import pandas as pd
import matplotlib.pyplot as plt
import plotly.graph_objects as go

# Data
import numpy as np

# Gaius Agent
from ia.gaius.agent_client import AgentClient
from ia.gaius.pvt.mongo_interface import MongoData, MongoResults
from ia.gaius.data_ops import Data, PreparedData
from ia.gaius.pvt.pvt_utils import init_emotive_on_node, plot_confusion_matrix, \
    compute_residual, compute_abs_residual, compute_squared_residual, smape, rmse,\
    f1_score, false_discovery_rate, false_positive_rate, update_accuracy, update_precision, is_notebook,\
    true_negative_rate, true_positive_rate, negative_predictive_value, false_negative_rate, false_omission_rate,\
    positive_likelihood_ratio, negative_likelihood_ratio, prevalence_threshold, threat_score,\
    init_emotive_polarity_results, check_answer_correctness, on_error_learn, plot_emotives_value_charts
from ia.gaius.prediction_models import hive_model_classification, average_emotives, prediction_ensemble_model_classification,\
    most_common_ensemble_model_classification, prediction_ensemble_modeled_emotives

logger = logging.getLogger(__name__)


class PVTAbortError(Exception):
    """Raised when PVT is aborted by Celery. Used to exit cleanly from nested test/train functions"""
    pass


class PVTConfigError(Exception):
    pass


class PVTMessage():
    """Wrapper for PVT socket messages to be sent during training and testing"""

    def __init__(self, status: str,
                 current_record: int,
                 total_record_count: int,
                 metrics: dict,
                 cur_test_num: int,
                 total_test_num: int,
                 test_id: str = None,
                 user_id: str = '',
                 test_type: str = 'default',
                 overall_metrics: dict = None):
        self.status = status
        self.current_record = current_record
        self.total_record_count = total_record_count
        self.metrics = metrics
        self.test_id = test_id
        self.user_id = user_id
        self.cur_test_num = cur_test_num
        self.total_test_num = total_test_num
        self.test_type = test_type
        self.overall_metrics = overall_metrics

    def toJSON(self):
        return json.loads(json.dumps({'status': self.status,
                                      'current_record': self.current_record,
                                      'total_record_count': self.total_record_count,
                                      'metrics': self.metrics,
                                      'overall_metrics': self.overall_metrics,
                                      'test_id': self.test_id,
                                      'user_id': self.user_id,
                                      'cur_test_num': self.cur_test_num,
                                      'total_test_num': self.total_test_num,
                                      'test_type': self.test_type
                                      }))


class PerformanceValidationTest():
    """
    Performance Validation Test (PVT) - Splits a GDF folder into training and testing sets.
    Based on the test type certain visualizations will be produced.

    Test types:

    - Classification
    - Emotive Value
    - Emotives Polarity
    """

    def __init__(self,
                 agent: AgentClient,
                 ingress_nodes: list,
                 query_nodes: list,
                 test_type: str,
                 test_count: int,
                 dataset: Union[Path, Data, MongoData, PreparedData],
                 test_prediction_strategy="continuous",
                 clear_all_memory_before_training: bool = True,
                 turn_prediction_off_during_training: bool = False,
                 learning_strategy: str = 'after_every',
                 shuffle: bool = False,
                 dataset_info: dict = None,
                 **kwargs: dict):
        """Initialize the PVT object with all required parameters for execution

        Args:
            agent (AgentClient): GAIuS Agent to use for trainings
            ingress_nodes (list): Ingress nodes for the GAIuS Agent (see :func:`ia.gaius.agent_client.AgentClient.set_query_nodes`)
            query_nodes (list): Query nodes for the GAIuS Agent (see :func:`ia.gaius.agent_client.AgentClient.set_query_nodes`)
            test_type (str): classification, emotives_value, or emotives_polarity
            test_count (int): Number of tests to run
            dataset (Union[Path, Data, MongoData, PreparedData]): Dataset path or object
            test_prediction_strategy (str, optional): when to learn new sequences. continuous -> learn during training and testing. noncontinuous -> learn only during training. Defaults to "continuous".
            clear_all_memory_before_training (bool, optional): Whether the GAIuS agent's memory should be cleared before each training. Defaults to True.
            turn_prediction_off_during_training (bool, optional): Whether predictions should be disabled during training to reduce computational load. Defaults to False.
            learning_strategy (str, optional): when learning is to be performed. Conforms to test_prediction_strategy. after_every -> learn after every sequence. on_error -> learn when agent guesses incorrectly
            shuffle (bool, optional): Whether dataset should be shuffled before each test iteration. Defaults to False.
            sio (_type_, optional): SocketIO object to emit information on. Defaults to None.
            task (_type_, optional): Celery details to emit information about. Defaults to None.
            user_id (str, optional): user_id to emit information to on SocketIO. Defaults to ''.
            mongo_db (pymongo.MongoClient, optional): MongoDB where dataset should be retrieved from
            dataset_info (dict, optional): information about how to retrieve dataset, used for MongoDB query
            test_id (str, optional): unique identifier to be sent with messages about this test. Also used for storing to mongodb
            test_configuration (dict, optional): dictionary storing additional metadata about test configuration, to be saved in mongodb with test results
            socket_channel (str, optional): SocketIO channel to broadcast results on. Defaults to 'pvt_status'
            QUIET (bool, optional): flag used to disable log output during PVT. Defaults to False
            DISABLE_TQDM (bool, optional): flag used to disable TQDM progress bars during PVT. Defaults to None (enabled)
            PLOT (bool, optional): flag used to determine whether to plot result graphs. Defaults to True
        """

        self.agent = agent
        self.ingress_nodes = ingress_nodes
        self.query_nodes = query_nodes
        self.test_count = test_count
        self.dataset = dataset

        self.dataset_percentage = kwargs.get('dataset_percentage', None)
        self.training_percentage = kwargs.get('training_percentage', None)
        self.shuffle = shuffle
        self.test_type = test_type
        self.clear_all_memory_before_training = clear_all_memory_before_training
        self.turn_prediction_off_during_training = turn_prediction_off_during_training
        self.test_prediction_strategy = test_prediction_strategy

        self.emotives_set = defaultdict(set)
        self.labels_set = defaultdict(set)
        self.predictions = None
        self.actuals = None
        self.emotive_value_results = None
        self.pvt_results = None

        self.dataset_info = dataset_info
        self.testing_log = []
        self.mongo_results = None
        self.learning_strategy = learning_strategy
        self.labels_counter = Counter()
        self.overall_labels_counter = Counter()
        self.testing_counter = Counter()
        self.training_counter = Counter()
        self.overall_training_counter = Counter()
        self.overall_testing_counter = Counter()
        self.predicted_class_statistics = defaultdict(Counter)
        self.overall_predicted_class_statistics = defaultdict(Counter)

        self.test_id: str = kwargs.get('test_id', None)
        self.user_id: str = kwargs.get('user_id', None)
        self.socket_channel = kwargs.get('socket_channel', 'pvt_status')
        self.QUIET: bool = kwargs.get('QUIET', False)
        self.DISABLE_TQDM: bool = kwargs.get('DISABLE_TQDM', None)
        self.sio = kwargs.get('sio', None)
        self.task = kwargs.get('task', None)
        self.mongo_db = kwargs.get('mongo_db', None)
        self.test_configuration = kwargs.get('test_configuration', {})
        self.results_filepath = kwargs.get('results_filepath', None)
        self.PLOT = kwargs.get('PLOT', True)

        self.overall_metrics = None
        self.overall_results = None
        self._original_dataset = self.dataset
        self.__validate_settings()

        # retrieve genes before test
        self.agent_genes = {}
        all_nodes = [node['name'] for node in self.agent.all_nodes]
        
        # Setting summarize single to False by default in order to handle multiply nodes topologies
        self.agent.set_summarize_for_single_node(False)

        self.agent_genes = self.agent.get_all_genes(nodes=all_nodes)
        self.test_configuration['initial_agent_genes'] = self.agent_genes
        self.test_configuration['genome'] = self.agent.genome.topology
        
        if isinstance(self._original_dataset, MongoData):
            self.mongo_results = MongoResults(mongo_db=self.mongo_db,
                                              result_collection_name=self.dataset_info['results_collection'],
                                              log_collection_name=self.dataset_info['logs_collection'],
                                              test_id=self.test_id,
                                              user_id=self.user_id,
                                              dataset_id=self.dataset_info['dataset_id'],
                                              test_configuration=self.test_configuration)
        elif (isinstance(self._original_dataset, Path) or isinstance(self._original_dataset, str)):
            self.dataset = Data(data_directories=[self.dataset])
        elif isinstance(self._original_dataset, PreparedData):
            pass
        elif isinstance(self._original_dataset, Data):
            pass

        if self.results_filepath is not None:
            if not os.path.exists(self.results_filepath):
                os.makedirs(self.results_filepath)
        # Show Agent status by Default
        self.agent.show_status()

        # Assign Ingress and Query Nodes
        self.agent.set_ingress_nodes(nodes=self.ingress_nodes)
        self.agent.set_query_nodes(nodes=self.query_nodes)



        logger.debug(f"test_counts = {self.test_count}")

    def __validate_settings(self):
        if self.test_prediction_strategy not in ['continuous', 'noncontinuous']:
            raise Exception(
                """
                Not a valid test prediction strategy. Please choose either 
                'continuous': learn the test sequence/answer after the agent has tried to make a prediction
                'noncontinuous': do not learn the test sequence.
                """
            )
        if self.learning_strategy not in ['after_every', 'on_error']:
            raise Exception(
                """Not a valid learning strategy. Please choose either
        'after_every': learn every sequence
        'on_error': learn the sequence only when the agent guesses incorrectly.
        """
            )
        if self.learning_strategy == 'on_error':
            if self.turn_prediction_off_during_training == True:
                raise Exception(
                    """When learning_strategy is 'on_error', predictions must be enabled during training.
                    """
                )
            if self.test_type not in ['classification', 'emotives_polarity']:
                raise Exception(
                    """When learning_strategy is 'on_error', test_type must be either 'classification' or 'emotives_polarity'
                    """
                )
        if not any([isinstance(self.dataset, ds_type) for ds_type in [Data, MongoData, PreparedData, Path, str]]):
            raise Exception(
                f'unknown type for dataset: {type(self.dataset)}')
        if self.test_type not in ['classification', 'emotives_value', 'emotives_polarity']:
            raise Exception(f'unknown value for test_type: {self.test_type}')

        if any([isinstance(self._original_dataset, ds_type) for ds_type in [Path, str]]):
            if (self.dataset_percentage == None) or (self.training_percentage is None):
                raise Exception(
                    "dataset_percentage and training_percentage must be provided when using filepath for dataset")

        return

    def prepare_datasets(self):
        
        percent_of_dataset_chosen = self.dataset.percent_of_dataset_chosen
        percent_reserved_for_training = self.dataset.percent_reserved_for_training
        if self.dataset_percentage is not None:
            percent_of_dataset_chosen = self.dataset_percentage
        if self.training_percentage is not None:
            percent_reserved_for_training = self.training_percentage

        self.dataset.prep(
            percent_of_dataset_chosen=percent_of_dataset_chosen,
            percent_reserved_for_training=percent_reserved_for_training,
            shuffle=self.shuffle
        )
        logger.debug(
            f"Length of Training Set = {len(self.dataset.train_sequences)}")
        logger.debug(
            f"Length of Testing Set  = {len(self.dataset.test_sequences)}")
        return

    def run_classification_pvt(self):
        for test_num in range(0, self.test_count):
            self.test_num = test_num
            logger.debug(f'Conducting Test # {test_num}')
            logger.debug('\n---------------------')

            self.prepare_datasets()

            if self.sio:  # pragma: no cover
                self.sio.emit(self.socket_channel,
                              PVTMessage(status='training',
                                         current_record=0,
                                         total_record_count=len(
                                             self.dataset.train_sequences),
                                         metrics={},
                                         overall_metrics={},
                                         cur_test_num=self.test_num,
                                         total_test_num=self.test_count-1,
                                         test_id=self.test_id,
                                         user_id=self.user_id,
                                         test_type=self.test_type).toJSON(),
                              to=self.user_id)
            try:

                self.train_agent()

                if len(self.dataset.test_sequences) == 0:
                    logger.debug(f'Complete!')
                    continue

                self.test_agent()

                for k, labels in self.labels_set.items():
                    self.labels_set[k] = set(
                        [label.rsplit('|', maxsplit=1)[-1] for label in labels])
                logger.debug('Getting Classification Metrics...')
                class_metrics_data_structures = get_classification_metrics(
                    labels_set=self.labels_set, this_test_log=self.testing_log[self.test_num], overall=False)
                self.overall_results = get_classification_metrics(
                    labels_set=self.labels_set, this_test_log=list(chain(*self.testing_log)), overall=True)
                self.pvt_results.append(
                    deepcopy(class_metrics_data_structures))

            except Exception as e:
                logger.error(
                    'error during training/testing phase of test, remediating database for failed test, then raising error')
                if self.mongo_results:
                    logger.info('about to remediate database')
                    self.mongo_results.deleteResults()
                    logger.info('remediated database')

                logger.debug(f'raising error {str(e)}')
                raise e

            try:
                if not isinstance(self.dataset, MongoData) and (self.PLOT == True):
                    logger.debug('Plotting Results...')
                    plot_confusion_matrix(test_num=test_num,
                                          class_metrics_data_structures=class_metrics_data_structures,
                                          results_dir=self.results_filepath)
            except Exception as e:
                logger.exception(
                    f'error plotting results from classification pvt: {str(e)}')
                pass

            response_dict = {'counter': self.labels_counter,
                             'pvt_results': self.pvt_results,
                             'overall_results': self.overall_results,
                             'final_agent_status': self.agent.show_status()}
            result_msg = PVTMessage(status='finished',
                                    current_record=0,
                                    total_record_count=0,
                                    metrics=response_dict,
                                    overall_metrics={},
                                    cur_test_num=self.test_num,
                                    total_test_num=self.test_count-1,
                                    test_id=self.test_id,
                                    user_id=self.user_id,
                                    test_type=self.test_type).toJSON()
            if self.sio:  # pragma: no cover
                self.sio.emit(self.socket_channel,
                              result_msg,
                              to=self.user_id)

        if self.mongo_results:
            self.mongo_results.saveResults(result_msg)
        return

    def run_emotive_value_pvt(self):
        self.pvt_results = []
        for test_num in range(0, self.test_count):
            self.test_num = test_num
            logger.debug(f'Conducting Test # {test_num}')
            logger.debug('\n---------------------\n')

            self.prepare_datasets()
            if self.sio:  # pragma: no cover
                self.sio.emit(self.socket_channel,
                              PVTMessage(status='training',
                                         current_record=0,
                                         total_record_count=len(
                                             self.dataset.train_sequences),
                                         metrics={},
                                         cur_test_num=self.test_num,
                                         total_test_num=self.test_count-1,
                                         test_id=self.test_id,
                                         user_id=self.user_id,
                                         test_type=self.test_type).toJSON(),
                              to=self.user_id)

            self.train_agent()

            if len(self.dataset.test_sequences) == 0:
                logger.debug(f'Complete!')
                return

            self.test_agent()

            self.emotive_value_results = get_emotives_value_metrics(
                emotives_set=self.emotives_set, this_test_log=self.testing_log[self.test_num], overall=False)
            self.overall_results = get_emotives_value_metrics(
                emotives_set=self.emotives_set, this_test_log=list(chain(*self.testing_log)), overall=True)
            self.pvt_results.append(deepcopy(self.emotive_value_results))
            logger.debug('Plotting Results...')

            # don't try to plot emotive values if we're working to save in a mongo database
            # (its probably running without a jupyter GUI)
            if self.PLOT:
                plot_emotives_value_charts(test_num=self.test_num,
                                           emotive_value_results=self.emotive_value_results,
                                           QUIET=self.QUIET,
                                           results_filepath=self.results_filepath)

        # send out finished socket message
        response_dict = {'counter': self.labels_counter,
                         'pvt_results': self.pvt_results,
                         'overall_results': self.overall_results,
                         'final_agent_status': self.agent.show_status()}

        final_msg = PVTMessage(status='finished',
                               current_record=0,
                               total_record_count=0,
                               metrics=response_dict,
                               overall_metrics={},
                               cur_test_num=self.test_num,
                               total_test_num=self.test_count-1,
                               test_id=self.test_id,
                               user_id=self.user_id,
                               test_type=self.test_type).toJSON()
        if self.sio:  # pragma: no cover
            self.sio.emit(self.socket_channel,
                          final_msg,
                          to=self.user_id)
        if self.mongo_results:
            self.mongo_results.saveResults(final_msg)
        return

    def run_emotive_polarity_pvt(self):
        self.pvt_results = []
        for test_num in range(0, self.test_count):
            self.test_num = test_num
            logger.debug(f'Conducting Test # {test_num}')

            self.prepare_datasets()

            if self.sio:  # pragma: no cover
                self.sio.emit(self.socket_channel,
                              PVTMessage(status='training',
                                         current_record=0,
                                         total_record_count=len(
                                             self.dataset.train_sequences),
                                         metrics={},
                                         overall_metrics={},
                                         cur_test_num=self.test_num,
                                         total_test_num=self.test_count-1,
                                         test_id=self.test_id,
                                         user_id=self.user_id,
                                         test_type=self.test_type).toJSON(),
                              to=self.user_id)

            logger.debug("Training Agent...")
            self.train_agent()

            if len(self.dataset.test_sequences) == 0:
                logger.debug(f'Complete!')
                return

            logger.debug("Testing Agent...")
            self.test_agent()

            logger.debug('Getting Emotives Polarity Metrics...')
            logger.debug('Saving results to pvt_results...')
            self.pvt_results.append(
                deepcopy(get_emotives_polarity_metrics(emotives_set=self.emotives_set,
                                                       this_test_log=self.testing_log[self.test_num],
                                                       overall=False)))
            self.overall_results = get_emotives_polarity_metrics(
                emotives_set=self.emotives_set,
                this_test_log=list(chain(*self.testing_log)),
                overall=True)

        # send out finished socket message
        response_dict = {'counter': self.labels_counter,
                         'pvt_results': self.pvt_results,
                         'overall_results': self.overall_results,
                         'final_agent_status': self.agent.show_status()}
        final_msg = PVTMessage(status='finished',
                               current_record=0,
                               total_record_count=0,
                               metrics=response_dict,
                               overall_metrics={},
                               cur_test_num=self.test_num,
                               total_test_num=self.test_count-1,
                               test_id=self.test_id,
                               user_id=self.user_id,
                               test_type=self.test_type).toJSON()
        if self.sio:  # pragma: no cover
            self.sio.emit(self.socket_channel,
                          final_msg, to=self.user_id)
        if self.mongo_results:
            self.mongo_results.saveResults(final_msg)
        return

    def conduct_pvt(self):
        """
        Function called to execute the PVT session. Determines test to run based on 'test_type' attribute

        Results from PVT is stored in the 'pvt_results' attribute

        .. note::

            A complete example is shown in the :ref:`Usage Examples <training reference>` document. Please see this doc for further information about how to conduct a PVT test

        """

        try:
            self.test_num = 0
            self.pvt_results = []
            self.testing_log = []

            self.overall_metrics = None

            if self.test_type in ['classification']:
                self.overall_labels_counter = Counter()
                self.overall_testing_counter = Counter()
                self.overall_training_counter = Counter()
                self.overall_predicted_class_statistics = defaultdict(Counter)
                self.overall_metrics = defaultdict(lambda: defaultdict(float))
            elif self.test_type in ['emotives_value']:
                self.overall_labels_counter = Counter()
                self.overall_testing_counter = Counter()
                self.overall_training_counter = Counter()
                self.overall_predicted_class_statistics = defaultdict(Counter)
                self.overall_metrics = defaultdict(
                    lambda: defaultdict(lambda: defaultdict(float)))
            elif self.test_type in ['emotives_polarity']:
                self.labels_counter = defaultdict(Counter)
                self.testing_counter = defaultdict(Counter)
                self.training_counter = defaultdict(Counter)
                self.overall_labels_counter = defaultdict(Counter)
                self.overall_testing_counter = defaultdict(Counter)
                self.overall_training_counter = defaultdict(Counter)
                self.predicted_class_statistics = defaultdict(
                    lambda: defaultdict(Counter))
                self.overall_predicted_class_statistics = defaultdict(
                    lambda: defaultdict(Counter))

                # init overall metrics for emotive polarity
                self.overall_metrics = defaultdict(
                    lambda: defaultdict(lambda: defaultdict(float)))
                self.overall_metrics['positive'] = defaultdict(
                    lambda: defaultdict(lambda: defaultdict(float)))
                self.overall_metrics['negative'] = defaultdict(
                    lambda: defaultdict(lambda: defaultdict(float)))
                self.overall_metrics['overall'] = defaultdict(
                    lambda: defaultdict(lambda: defaultdict(float)))

            # Validate Test Type
            if self.test_type == 'classification':
                logger.debug("Conducting Classification PVT...\n")
                self.run_classification_pvt()

            elif self.test_type == 'emotives_value':
                logger.debug("Conducting Emotives Value PVT...\n")
                self.run_emotive_value_pvt()

            elif self.test_type == 'emotives_polarity':
                logger.debug("Conducting Emotives Polarity PVT...\n")
                self.run_emotive_polarity_pvt()

            else:
                raise Exception(
                    """
                    Please choose one of the test type:
                    - classification
                    - emotives_value
                    - emotives_polarity

                    ex.
                    --> pvt.test_type='emotives_value'
                    then, retry
                    --> pvt.conduct_pvt()
                    """
                )
        except Exception as e:
            logger.exception(
                f'failed to conduct PVT test, test_type={self.test_type}: {str(e)}')
            raise e

        # convert defaultdict to normal dict by dumping and loading pvt results
        self.pvt_results = json.loads(json.dumps(self.pvt_results))
        self.overall_metrics = json.loads(json.dumps(self.overall_metrics))

    def train_agent(self):
        """
        Takes a training set of gdf files, and then trains an agent on those records.
        The user can turn prediction off if the topology doesn't have abstractions
        where prediction is needed to propagate data through the topology.
        """
        # Initialize
        if self.clear_all_memory_before_training is True:
            logger.debug('Clearing memory of selected ingress nodes...')
            self.agent.clear_all_memory()

        self.labels_set.clear()
        self.labels_counter.clear()
        self.training_counter.clear()
        self.testing_counter.clear()
        self.emotives_set.clear()
        self.predicted_class_statistics.clear()

        # Train Agent
        if self.turn_prediction_off_during_training is True:
            self.agent.stop_predicting(nodes=self.query_nodes)
        else:
            self.agent.start_predicting(nodes=self.query_nodes)

        logger.debug('Preparing to train agent...')

        train_seq_len = len(self.dataset.train_sequences)

        train_metrics, train_progress_bar = self._setup_training()

        for j, _ in enumerate(train_progress_bar):

            if j % 10 == 0:
                if self.task:  # pragma: no cover (testing disabled for Celery code (used by Lab))
                    if self.task.is_aborted():
                        self.abort_test_remediation(
                            current_record=j, record_count=train_seq_len)
                        return

            sequence = self._extract_training_sequence(j)

            current_labels = None
            record_emotive_set = None

            # update label/emotive counters
            if self.test_type == 'classification':

                train_metrics, current_labels = self._train_classification(
                    sequence=sequence, train_metrics=train_metrics)

            elif self.test_type in ['emotives_value', 'emotives_polarity']:

                train_metrics, record_emotive_set = self._train_emotives(
                    sequence=sequence, train_metrics=train_metrics)

            self._apply_learning_strategy(
                sequence=sequence, current_labels=current_labels, record_emotive_set=record_emotive_set)
            self.overall_metrics.update({'counter': self.overall_labels_counter,
                                         'training_counter': self.overall_training_counter})

            training_msg = PVTMessage(status='training',
                                      current_record=j + 1,
                                      total_record_count=train_seq_len,
                                      metrics=train_metrics,
                                      overall_metrics=self.overall_metrics,
                                      cur_test_num=self.test_num,
                                      total_test_num=self.test_count-1,
                                      test_id=self.test_id,
                                      user_id=self.user_id,
                                      test_type=self.test_type)

            self.store_train_record(
                test_num=self.test_num, record=training_msg)

        train_progress_bar.reset()
        logger.debug('Finished training agent!')

    def _apply_learning_strategy(self, sequence, current_labels, record_emotive_set):

        if self.learning_strategy == 'on_error':

            predictions = self.agent.get_predictions()
            self.agent.observe(data=sequence[-1], nodes=self.ingress_nodes)
            if self.test_type == 'classification':
                on_error_learn(agent=self.agent, test_type=self.test_type, query_nodes=self.query_nodes,
                               predictions=predictions, current_labels=current_labels, record_emotive_set=None)
            elif self.test_type == 'emotives_polarity':
                on_error_learn(agent=self.agent, test_type=self.test_type, query_nodes=self.query_nodes,
                               predictions=predictions, current_labels=None, record_emotive_set=record_emotive_set)

        elif self.learning_strategy == 'after_every':
            if self.test_type == 'classification':
                self.agent.observe(
                    data=sequence[-1], nodes=self.ingress_nodes)
            self.agent.learn(nodes=self.ingress_nodes)

    def _setup_training(self):

        train_metrics = {}
        if self.test_type == 'classification':
            train_metrics = {'counter': self.labels_counter,
                             'training_counter': self.training_counter,
                             'testing_counter': self.testing_counter,
                             'predicted_class_statistics': self.predicted_class_statistics}
        elif self.test_type == 'emotives_polarity':
            train_metrics = {'counter': self.labels_counter,
                             'training_counter': self.training_counter,
                             'testing_counter': self.testing_counter}
        elif self.test_type == 'emotives_value':

            train_metrics = {'counter': self.labels_counter,
                             'training_counter': self.training_counter,
                             'testing_counter': self.testing_counter}

        train_progress_bar = tqdm(self.dataset.train_sequences,
                                  bar_format="{l_bar}{bar} {n_fmt}/{total_fmt} [{remaining} {rate_fmt}{postfix}]",
                                  disable=self.DISABLE_TQDM,
                                  leave=True,
                                  unit=' records')
        train_progress_bar.set_description(f'Training (Test #{self.test_num})')
        train_progress_bar.unit = ' records'

        return train_metrics, train_progress_bar

    def _extract_training_sequence(self, idx):
        if isinstance(self.dataset, PreparedData):
            sequence = self.dataset.train_sequences[idx]
        elif isinstance(self.dataset, Data):
            with open(self.dataset.train_sequences[idx], "r") as sequence_file:
                sequence = sequence_file.readlines()
                sequence = [json.loads(d) for d in sequence]
        elif isinstance(self.dataset, MongoData):
            sequence = self.dataset.getSequence(
                self.dataset.train_sequences[idx])

        return sequence

    def _train_emotives(self, sequence, train_metrics):

        # observe training sequence
        for event in sequence:
            self.agent.observe(data=event, nodes=self.ingress_nodes)
            for node in self.ingress_nodes:
                percept_emotives = list(self.agent.get_percept_data()[
                                        node]['emotives'].keys())
                self.emotives_set[node].update(percept_emotives)

        # compute emotive information in current sequence, for emotive tests
        if self.test_type == 'emotives_polarity':
            record_emotive_set = dict()
            for event in sequence:
                record_emotive_set.update(event['emotives'])
        elif self.test_type == 'emotives_value':
            record_emotive_set = set()
            for event in sequence:
                record_emotive_set.update(
                    list(event['emotives'].keys()))

        self.update_training_counters(
            current_labels=record_emotive_set)

        train_metrics.update(
            {'actual': self.sum_sequence_emotives(sequence)})
        self.overall_metrics.update(
            {'actual': self.sum_sequence_emotives(sequence)})

        return train_metrics, record_emotive_set

    def _train_classification(self, sequence, train_metrics):
        for event in sequence[:-1]:
            self.agent.observe(data=event, nodes=self.ingress_nodes)
        current_labels = [label.rsplit(
            '|', maxsplit=1)[-1] for label in sequence[-1]['strings']]
        for node in self.ingress_nodes:
            self.labels_set[node].update(current_labels)

        self.update_training_counters(current_labels=current_labels)

        train_metrics.update({'actual': current_labels})
        self.overall_metrics.update({'actual': current_labels})

        return train_metrics, current_labels

    def test_agent(self):
        """
        Test agent on dataset test sequences provided in self.dataset.test_sequences
        """
        # Start Testing
        self.agent.start_predicting(nodes=self.query_nodes)
        self.predictions = []
        self.actuals = []

        self.testing_log.append([])

        test_seq_len = len(self.dataset.test_sequences)
        if test_seq_len == 0:
            logger.debug('length of testing sequences is 0... returning\n')
            return

        test_step_info, test_progress_bar = self._setup_testing()

        for k, _ in enumerate(test_progress_bar):

            if k % 10 == 0:
                if self.task:  # pragma: no cover (testing disabled for Celery code (used by Lab))
                    if self.task.is_aborted():
                        self.abort_test_remediation(
                            current_record=k, record_count=test_seq_len)
                        return

            sequence = self._extract_testing_sequence(k)

            current_labels = None
            record_emotive_set = None

            self.agent.clear_wm(nodes=self.ingress_nodes)
            if self.test_type == 'classification':
                test_step_info, current_labels = self._test_classification(
                    sequence=sequence, test_step_info=test_step_info, idx=k)

            elif self.test_type in ['emotives_value', 'emotives_polarity']:
                test_step_info, record_emotive_set = self._test_emotives(
                    sequence=sequence, test_step_info=test_step_info, idx=k)

            if self.test_prediction_strategy == 'continuous':
                self._apply_testing_strategy(
                    current_labels=current_labels, record_emotive_set=record_emotive_set)

            # prepare test step message
            test_step_msg = PVTMessage(status='testing',
                                       current_record=k + 1,
                                       total_record_count=test_seq_len,
                                       metrics=test_step_info,
                                       overall_metrics=self.overall_metrics,
                                       cur_test_num=self.test_num,
                                       total_test_num=self.test_count-1,
                                       test_id=self.test_id,
                                       user_id=self.user_id,
                                       test_type=self.test_type)

            self.store_train_record(test_num=self.test_num,
                                    record=test_step_msg)

        test_progress_bar.reset()
        return

    def _apply_testing_strategy(self, current_labels, record_emotive_set):

        if self.learning_strategy == 'on_error':
            if self.test_type == 'classification':
                on_error_learn(agent=self.agent, test_type=self.test_type, query_nodes=self.query_nodes,
                               predictions=self.predictions[-1], current_labels=current_labels, record_emotive_set=None)
            elif self.test_type == 'emotives_polarity':
                on_error_learn(agent=self.agent, test_type=self.test_type, query_nodes=self.query_nodes,
                               predictions=self.predictions[-1], current_labels=None, record_emotive_set=record_emotive_set)

        elif self.learning_strategy == 'after_every':
            self.agent.learn(nodes=self.ingress_nodes)

    def _test_emotives(self, sequence, test_step_info, idx):

        if self.test_type == 'emotives_value':
            record_emotive_set = set()
            for event in sequence:
                self.agent.observe(
                    data=event, nodes=self.ingress_nodes)
                record_emotive_set.update(
                    list(event['emotives'].keys()))
                for node in self.ingress_nodes:
                    self.emotives_set[node].update(
                        list(self.agent.get_percept_data()[node]['emotives'].keys()))
        elif self.test_type == 'emotives_polarity':
            record_emotive_set = dict()
            for event in sequence:
                self.agent.observe(
                    data=event, nodes=self.ingress_nodes)
                record_emotive_set.update(event['emotives'])
                for node in self.ingress_nodes:
                    self.emotives_set[node].update(
                        list(self.agent.get_percept_data()[node]['emotives'].keys()))

        # update counters with emotives from testing record
        self.update_testing_counters(current_labels=record_emotive_set)

        # get and store predictions after observing events
        self.predictions.append(
            self.agent.get_predictions(nodes=self.query_nodes))
        # store answers in a separate list for evaluation
        self.actuals.append(self.sum_sequence_emotives(sequence))

        pred_dict = {node: prediction_ensemble_modeled_emotives(
            self.predictions[idx][node]) for node in self.query_nodes}

        # compute hive prediction for time idx
        pred_dict['hive'] = average_emotives(list(pred_dict.values()))

        self.update_predicted_class_statistics(pred_dict=pred_dict)

        test_step_info.update({'idx': idx,
                               'predicted': pred_dict,
                               'actual': self.actuals[-1],
                               'counter': self.labels_counter,
                               'training_counter': self.training_counter,
                               'testing_counter': self.testing_counter,
                               'predicted_class_statistics': self.predicted_class_statistics})
        test_step_info = compute_incidental_probabilities(
            test_step_info=test_step_info, test_type=self.test_type)
        self.overall_metrics.update({'predicted': pred_dict,
                                     'actual': self.actuals[-1],
                                     'counter': self.overall_labels_counter,
                                     'training_counter': self.overall_training_counter,
                                     'testing_counter': self.overall_testing_counter,
                                     'predicted_class_statistics': self.overall_predicted_class_statistics})
        self.overall_metrics['idx'] = self.overall_metrics.get(
            'idx', idx-1) + 1
        # doing compute incidental probabilities a second time, for the overall metrics
        self.overall_metrics = compute_incidental_probabilities(
            test_step_info=self.overall_metrics, test_type=self.test_type)

        return test_step_info, record_emotive_set

    def _test_classification(self, sequence, test_step_info, idx):

        for event in sequence[:-1]:
            self.agent.observe(data=event, nodes=self.ingress_nodes)

        # get and store predictions after observing events
        self.predictions.append(
            self.agent.get_predictions(nodes=self.query_nodes))

        # store answers in a separate list for evaluation
        current_labels = [label.rsplit('|', maxsplit=1)[-1]
                          for label in sequence[-1]['strings']]

        self.actuals.append(deepcopy(current_labels))

        for node in self.ingress_nodes:
            self.labels_set[node].update(current_labels)

        # persists across multiple runs
        self.update_testing_counters(current_labels)

        # get predicted classification on the fly, so we can save to mongo individually
        pred_dict = {node: most_common_ensemble_model_classification(
            self.predictions[idx][node]) for node in self.query_nodes}
        pred_dict['hive'] = hive_model_classification(
            self.predictions[idx])
        if pred_dict['hive'] is not None:
            pred_dict['hive'] = pred_dict['hive'].most_common(1)[0][0]

        self.update_predicted_class_statistics(pred_dict=pred_dict)
        test_step_info.update({'idx': idx,
                               'predicted': pred_dict,
                               'actual': self.actuals[idx],
                               'counter': self.labels_counter,
                               'training_counter': self.training_counter,
                               'testing_counter': self.testing_counter,
                               'predicted_class_statistics': self.predicted_class_statistics})
        test_step_info = compute_incidental_probabilities(
            test_step_info=test_step_info, test_type=self.test_type)
        self.overall_metrics.update({'predicted': pred_dict,
                                     'actual': self.actuals[idx],
                                     'counter': self.overall_labels_counter,
                                     'training_counter': self.overall_training_counter,
                                     'testing_counter': self.overall_testing_counter,
                                     'predicted_class_statistics': self.overall_predicted_class_statistics})
        self.overall_metrics['idx'] = self.overall_metrics.get(
            'idx', idx-1) + 1
        # doing compute incidental probabilities a second time, for the overall metrics
        self.overall_metrics = compute_incidental_probabilities(
            test_step_info=self.overall_metrics, test_type=self.test_type)

        # observe answer
        self.agent.observe(sequence[-1], nodes=self.ingress_nodes)

        return test_step_info, current_labels

    def _setup_testing(self):
        if self.test_type == 'classification':
            test_step_info = defaultdict(lambda: defaultdict(float))

        elif self.test_type == 'emotives_polarity':
            test_step_info = defaultdict(
                lambda: defaultdict(lambda: defaultdict(float)))
            test_step_info['positive'] = defaultdict(
                lambda: defaultdict(lambda: defaultdict(float)))
            test_step_info['negative'] = defaultdict(
                lambda: defaultdict(lambda: defaultdict(float)))
            test_step_info['overall'] = defaultdict(
                lambda: defaultdict(lambda: defaultdict(float)))

        elif self.test_type == 'emotives_value':
            test_step_info = defaultdict(
                lambda: defaultdict(lambda: defaultdict(float)))

        test_progress_bar = tqdm(self.dataset.test_sequences,
                                 bar_format="{l_bar}{bar} {n_fmt}/{total_fmt} [{remaining} {rate_fmt}{postfix}]",
                                 disable=self.DISABLE_TQDM,
                                 leave=True,
                                 unit=' records')
        test_progress_bar.set_description(f'Testing (Test #{self.test_num})')

        return test_step_info, test_progress_bar

    def _extract_testing_sequence(self, idx):
        if isinstance(self.dataset, PreparedData):
            sequence = self.dataset.test_sequences[idx]
        elif isinstance(self.dataset, Data):
            with open(self.dataset.test_sequences[idx], "r") as sequence_file:
                sequence = sequence_file.readlines()
                sequence = [json.loads(d) for d in sequence]
        elif isinstance(self.dataset, MongoData):
            sequence = self.dataset.getSequence(
                self.dataset.test_sequences[idx])


        return sequence

    def update_training_counters(self, current_labels):
        if self.test_type in ['classification', 'emotives_value']:
            self.labels_counter.update(current_labels)
            self.training_counter.update(current_labels)
            self.overall_labels_counter.update(current_labels)
            self.overall_training_counter.update(current_labels)
        elif self.test_type in ['emotives_polarity']:
            for emo, val in current_labels.items():
                emotive_sign = np.sign(val)
                if emotive_sign == 1:
                    emotive_label_types = {
                        'positive': 1, 'overall': 1, 'negative': 0}
                elif emotive_sign == -1:
                    emotive_label_types = {
                        'positive': 0, 'overall': 1, 'negative': 1}
                self.labels_counter[emo].update(emotive_label_types)
                self.training_counter[emo].update(emotive_label_types)
                self.overall_labels_counter[emo].update(emotive_label_types)
                self.overall_training_counter[emo].update(emotive_label_types)

    def update_testing_counters(self, current_labels):
        if self.test_type in ['classification', 'emotives_value']:
            self.labels_counter.update(current_labels)
            self.testing_counter.update(current_labels)
            self.overall_labels_counter.update(current_labels)
            self.overall_testing_counter.update(current_labels)
        elif self.test_type in ['emotives_polarity']:
            for emo, val in current_labels.items():
                emotive_sign = np.sign(val)
                if emotive_sign == 1:
                    emotive_label_types = {
                        'positive': 1, 'overall': 1, 'negative': 0}
                elif emotive_sign == -1:
                    emotive_label_types = {
                        'positive': 0, 'overall': 1, 'negative': 1}
                self.labels_counter[emo].update(emotive_label_types)
                self.testing_counter[emo].update(emotive_label_types)
                self.overall_labels_counter[emo].update(emotive_label_types)
                self.overall_testing_counter[emo].update(emotive_label_types)

    def update_predicted_class_statistics(self, pred_dict):
        """Update predicted class statistics based on test_type
        and provided dictionary of node predictions

        Args:
            pred_dict (dict): dict of node predictions
        """
        if self.test_type == 'classification':
            for key, val in pred_dict.items():
                self.predicted_class_statistics[key].update([val])
                self.overall_predicted_class_statistics[key].update([val])

        elif self.test_type == 'emotives_value':
            for key, modeled_emotives_dict in pred_dict.items():
                for emo, emotive_value in modeled_emotives_dict.items():

                    self.predicted_class_statistics[key].update([emo])
                    self.overall_predicted_class_statistics[key].update([
                                                                        emo])

        elif self.test_type == 'emotives_polarity':
            for key, modeled_emotives_dict in pred_dict.items():
                for emo, emotive_value in modeled_emotives_dict.items():
                    emotive_sign = np.sign(emotive_value)
                    if emotive_sign == 1:
                        emotive_label_types = {
                            'positive': 1, 'overall': 1, 'negative': 0}
                    elif emotive_sign == -1:
                        emotive_label_types = {
                            'positive': 0, 'overall': 1, 'negative': 1}
                    self.predicted_class_statistics[key][emo].update(
                        emotive_label_types)
                    self.overall_predicted_class_statistics[key][emo].update(
                        emotive_label_types)

    def sum_sequence_emotives(self, sequence):
        """
        Sums all emotive values
        """
        emotives_seq = [event['emotives']
                        for event in sequence if event['emotives']]
        return dict(functools.reduce(operator.add, map(Counter, emotives_seq)))

    def abort_test_remediation(self, current_record, record_count):  # pragma: no cover (testing disabled for Celery code (used by Lab))
        logger.info(
            f'about to abort {self.task.request.id =}, {self.test_id=}')
        if self.sio:  # pragma: no cover
            logger.info('Sending abort message')
            abort_msg = PVTMessage(status='aborted',
                                   current_record=current_record + 1,
                                   total_record_count=record_count,
                                   metrics={},
                                   overall_metrics={},
                                   cur_test_num=self.test_num,
                                   total_test_num=self.test_count-1,
                                   test_id=self.test_id,
                                   user_id=self.user_id,
                                   test_type=self.test_type)
            self.sio.emit(self.socket_channel,
                          abort_msg.toJSON(), to=self.user_id)
        if self.mongo_results:
            logger.info('cleaning up MongoDB')
            self.mongo_results.deleteResults()

        raise PVTAbortError(
            f"Aborting Test, at record {current_record} of {record_count}")

    def store_train_record(self, test_num, record: PVTMessage):

        if record.status == 'testing':
            self.testing_log[test_num].append(deepcopy(record.toJSON()))

        # insert into test_log in mongo, if using mongodb
        if self.mongo_results:
            self.mongo_results.addLogRecord(
                type=record.status, record=deepcopy(record.toJSON()))

        # emit socketIO message
        if self.sio:  # pragma: no cover
            self.sio.emit(self.socket_channel, deepcopy(
                record.toJSON()), to=self.user_id)


def get_classification_metrics(labels_set, this_test_log, overall=False):
    """
    Builds classification final results structure for each node
    """

    class_metrics = defaultdict(lambda: defaultdict(list))
    actuals = []

    if overall:
        metrics_key = 'overall_metrics'
    else:
        metrics_key = 'metrics'

    for record in this_test_log:
        record_metrics = record[metrics_key]
        actuals.append(record_metrics['actual'])
        for node in record_metrics['predicted'].keys():
            class_metrics[node]['predictions'].append(
                record_metrics['predicted'][node])
    last_test_record = this_test_log[-1]
    for node in class_metrics:
        class_metrics[node]['actuals'] = actuals
        class_metrics[node]['metrics'] = {}
        node_test_metrics = last_test_record[metrics_key]
        class_metrics[node]['metrics']['training_counter'] = node_test_metrics['training_counter']
        class_metrics[node]['metrics']['testing_counter'] = node_test_metrics['testing_counter']
        class_metrics[node]['metrics']['counter'] = node_test_metrics['counter']
        for metric, metric_values in node_test_metrics.items():
            if metric in ['training_counter', 'testing_counter', 'counter', 'idx', 'predicted', 'actual']:
                continue
            class_metrics[node]['metrics'][metric] = metric_values[node]
        if node == 'hive':
            continue
        class_metrics[node]['labels'] = list(
            set(list(labels_set[node]) + [None]))

    hive_label_set = set()
    label_set: set
    for label_set in labels_set.values():
        hive_label_set.update(label_set)
    class_metrics['hive']['labels'] = list(hive_label_set)

    return class_metrics


def get_emotives_value_metrics(emotives_set, this_test_log, overall=False):
    """
    Builds emotives value data structures for each node
    """
    # Build an emotives Metric Data Structure
    results = defaultdict(lambda: defaultdict(dict))
    # generate hive emotive set
    hive_emotive_set = set()
    label_set: set
    for label_set in emotives_set.values():
        hive_emotive_set.update(label_set)
    emotives_set['hive'] = hive_emotive_set

    if overall:
        metrics_key = 'overall_metrics'
    else:
        metrics_key = 'metrics'

    # init metrics structure for each node, emotive
    for node in emotives_set:
        for emotive in emotives_set[node]:
            results[node][emotive] = {}
            results[node][emotive]['actuals'] = []
            results[node][emotive]['predictions'] = []
            results[node][emotive]['residuals'] = []
            results[node][emotive]['abs_residuals'] = []
            results[node][emotive]['squared_residuals'] = []

    for record in this_test_log:
        current_metrics = record[metrics_key]
        for node, node_predicted_emotives in current_metrics['predicted'].items():
            for emotive, predicted_value in node_predicted_emotives.items():
                results[node][emotive]['predictions'].append(
                    predicted_value)
                results[node][emotive]['actuals'].append(
                    current_metrics['actual'][emotive])
                results[node][emotive]['residuals'].append(
                    current_metrics['residuals'][node][emotive])
                results[node][emotive]['abs_residuals'].append(
                    current_metrics['abs_residuals'][node][emotive])
                results[node][emotive]['squared_residuals'].append(
                    current_metrics['squared_residuals'][node][emotive])

    last_test_record = this_test_log[-1]
    test_metrics = last_test_record[metrics_key]
    for node in results:
        for emotive in emotives_set[node]:

            results[node][emotive]['metrics'] = {}
            results[node][emotive]['metrics']['response_counts'] = test_metrics['response_counts'][node][emotive]
            results[node][emotive]['metrics']['response_percentage'] = test_metrics['response_percentage'][node][emotive]
            results[node][emotive]['metrics']['unknown_percentage'] = test_metrics['unknown_percentage'][node][emotive]
            results[node][emotive]['metrics']['response_percentage'] = test_metrics['response_percentage'][node][emotive]
            results[node][emotive]['metrics']['counter'] = test_metrics['counter'][emotive]
            results[node][emotive]['metrics']['training_counter'] = test_metrics['training_counter'][emotive]
            results[node][emotive]['metrics']['testing_counter'] = test_metrics['testing_counter'][emotive]
            if (node in test_metrics['rmse']) and (emotive in test_metrics['rmse'][node]):
                results[node][emotive]['metrics']['rmse'] = test_metrics['rmse'][node][emotive]
                results[node][emotive]['metrics']['smape'] = test_metrics['smape'][node][emotive]
                results[node][emotive]['metrics']['1-smape'] = test_metrics['1-smape'][node][emotive]
            else:
                results[node][emotive]['metrics']['rmse'] = np.nan
                results[node][emotive]['metrics']['smape'] = np.nan
                results[node][emotive]['metrics']['1-smape'] = np.nan

    return results


def get_emotives_polarity_metrics(emotives_set, this_test_log, overall=False):
    """
    Builds emotives polarity data structures for each node
    """

    if overall:
        metrics_key = 'overall_metrics'
    else:
        metrics_key = 'metrics'

    template_dict = {'predictions': [],
                     'actuals': [],
                     'metrics': init_emotive_polarity_results(),
                     'predicted_class_statistics': {}
                     }
    if len(this_test_log) == 0:
        return {}
    # lets flip the dictionary so that it is organized per node instead of per metric
    raw_test_results = deepcopy(this_test_log[-1][metrics_key])

    flattened_emotive_set = set(
        chain(*[list(item) for item in emotives_set.values()]))
    hive_emotives_set = flattened_emotive_set
    emotive_polarity_results = {k: {i: deepcopy(template_dict) for i in v}
                                for k, v in emotives_set.items()}
    emotive_polarity_results['hive'] = {emo: deepcopy(template_dict)
                                        for emo in hive_emotives_set}
    for metric_type, metric_info in raw_test_results.items():
        if metric_type == 'predicted_class_statistics':
            for node, info in metric_info.items():
                for emotive, emotive_pcs in info.items():
                    emotive_polarity_results[node][emotive]['predicted_class_statistics'] = emotive_pcs
        elif metric_type not in ['overall', 'positive', 'negative']:
            continue
        for k, v in metric_info.items():
            if k not in ['true_positive',
                         'false_positive',
                         'true_negative',
                         'false_negative',
                         'unknown_percentage',
                         'response_percentage',
                         'response_counts',
                         'accuracy',
                         'precision',
                         'FPR',
                         'FDR',
                         'TNR',
                         'TPR',
                         'NPV',
                         'FNR',
                         'FOR',
                         'LR+',
                         'LR-',
                         'PT',
                         'TS']:
                continue

            for node, info in v.items():
                for emotive, emotive_data in info.items():
                    if emotive in emotive_polarity_results[node]:
                        emotive_polarity_results[node][emotive]['metrics'][metric_type][k] = deepcopy(
                            emotive_data)

                        emotive_polarity_results[node][emotive]['metrics'][metric_type][
                            'training_counter'] = raw_test_results['training_counter'][emotive][metric_type]
                        emotive_polarity_results[node][emotive]['metrics'][metric_type][
                            'testing_counter'] = raw_test_results['testing_counter'][emotive][metric_type]
                        emotive_polarity_results[node][emotive]['metrics'][metric_type][
                            'counter'] = raw_test_results['counter'][emotive][metric_type]

    for record in this_test_log:
        metrics = record[metrics_key]
        for emotive, val in metrics['actual'].items():
            for node, node_emotive_set in emotives_set.items():
                if emotive in node_emotive_set:
                    emotive_polarity_results[node][emotive]['actuals'].append(
                        val)
        for node, emotive_dict in metrics['predicted'].items():
            for emotive, val in emotive_dict.items():
                emotive_polarity_results[node][emotive]['predictions'].append(
                    val)
    return emotive_polarity_results


def compute_incidental_probabilities(test_step_info: dict, test_type: str):
    """Keep track of how well each node is doing during the testing phase. To be used for live visualizations

    Args:
        test_step_info (dict, required): Dictionary containing information about the current predicted, actual answers, and other related metrics (e.g. precision, unknowns, residuals, response rate, etc.)

    Returns:
        dict: updated test_step_info with the statistics for the current timestep
    """
    idx = test_step_info['idx']

    if test_type == 'classification':

        for k in test_step_info['predicted'].keys():
            if test_step_info['predicted'][k] != None:
                test_step_info['response_counts'][k] += 1

                if test_step_info['predicted'][k] in test_step_info['actual']:
                    test_step_info['true_positive'][k] += 1
                    # touch key in case it hasn't been initialized yet
                    test_step_info['false_positive'][k] += 0
                else:
                    test_step_info['false_positive'][k] += 1
                    # touch key in case it hasn't been initialized yet
                    test_step_info['true_positive'][k] += 0
            else:
                # touch key in case it hasn't been initialized yet
                test_step_info['response_counts'][k] += 0
                test_step_info['true_positive'][k] += 0
                test_step_info['false_positive'][k] += 0

            test_step_info['precision'][k] = update_precision(tp=test_step_info['true_positive'][k],
                                                              tn=0,
                                                              response_count=test_step_info['response_counts'][k])

            test_step_info['f1'][k] = f1_score(tp=test_step_info['true_positive'][k],
                                               fp=test_step_info['false_positive'][k],
                                               fn=0)

            test_step_info['accuracy'][k] = update_accuracy(tp=test_step_info['true_positive'][k],
                                                            tn=0,
                                                            overall_count=idx+1)
            # (test_step_info['true_positive'][k] / (idx + 1)) * 100
            test_step_info['response_percentage'][k] = (
                test_step_info['response_counts'][k] / (idx + 1)) * 100
            test_step_info['unknown_percentage'][k] = 100 - \
                test_step_info['response_percentage'][k]

            tp = test_step_info['true_positive'][k]
            tn = test_step_info['true_negative'][k]
            fp = test_step_info['false_positive'][k]
            fn = test_step_info['false_negative'][k]
            test_step_info['FPR'][k] = false_positive_rate(fp=fp, tn=tn)
            test_step_info['FDR'][k] = false_discovery_rate(tp=tp, fp=fp)
            test_step_info['TNR'][k] = true_negative_rate(tn=tn, fp=fp)
            test_step_info['TPR'][k] = true_positive_rate(tp=tp, fn=fn)
            test_step_info['NPV'][k] = negative_predictive_value(tn=tn, fn=fn)
            test_step_info['FNR'][k] = false_negative_rate(fn=fn, tp=tp)
            test_step_info['FOR'][k] = false_omission_rate(fn=fn, tn=tn)
            test_step_info['LR+'][k] = positive_likelihood_ratio(
                tp=tp, fp=fp, tn=tn, fn=fn)
            test_step_info['LR-'][k] = negative_likelihood_ratio(
                tp=tp, fp=fp, tn=tn, fn=fn)
            test_step_info['PT'][k] = prevalence_threshold(
                tp=tp, fp=fp, tn=tn, fn=fn)
            test_step_info['TS'][k] = threat_score(tp=tp, fp=fp, fn=fn)

    elif test_type == 'emotives_polarity':

        for k in test_step_info['predicted'].keys():
            for emotive in test_step_info['actual'].keys():
                actual_sign = np.sign(test_step_info['actual'][emotive])
                if actual_sign == 1:
                    actual_value_type = 'positive'
                elif actual_sign == -1:
                    actual_value_type = 'negative'
                if actual_sign == 0:
                    raise Exception(
                        f'Zero value found in polarity test at idx {idx}')

                # catch new emotives, not yet seen on node {k}
                if emotive not in test_step_info['overall']['true_positive'][k].keys():
                    init_emotive_on_node(
                        emotive=emotive, test_step_info=test_step_info, node=k)
                for val_type in [actual_value_type, 'overall']:
                    test_step_info[val_type]['testing_counter'][k][emotive] += 1
                if emotive in test_step_info['predicted'][k].keys():
                    pred_sign = np.sign(
                        test_step_info['predicted'][k][emotive])

                    for val_type in [actual_value_type, 'overall']:
                        # If predicted value non-zero
                        if bool(pred_sign):
                            test_step_info[val_type]['response_counts'][k][emotive] += 1

                        # True positive (correct)
                        if actual_sign > 0 and pred_sign > 0:
                            test_step_info[val_type]['true_positive'][k][emotive] += 1

                        # True Negative (correct)
                        elif actual_sign < 0 and pred_sign < 0:
                            test_step_info[val_type]['true_negative'][k][emotive] += 1

                        # False positive (incorrect)
                        elif actual_sign < 0 and not pred_sign < 0:
                            test_step_info[val_type]['false_positive'][k][emotive] += 1

                        # False negative (incorrect)
                        elif actual_sign > 0 and not pred_sign > 0:
                            test_step_info[val_type]['false_negative'][k][emotive] += 1

                        # Calculate precision value
                        tp = test_step_info[val_type]['true_positive'][k][emotive]
                        tn = test_step_info[val_type]['true_negative'][k][emotive]
                        fp = test_step_info[val_type]['false_positive'][k][emotive]
                        fn = test_step_info[val_type]['false_negative'][k][emotive]

                        test_step_info[val_type]['precision'][k][emotive] = update_precision(tp=tp,
                                                                                             tn=tn,
                                                                                             response_count=test_step_info[val_type]['response_counts'][k][emotive])

                        test_step_info[val_type]['accuracy'][k][emotive] = update_accuracy(tp=tp,
                                                                                           tn=tn,
                                                                                           overall_count=test_step_info[val_type]['testing_counter'][k][emotive])
                        test_step_info[val_type]['FPR'][k][emotive] = false_positive_rate(
                            fp=fp, tn=tn)
                        test_step_info[val_type]['FDR'][k][emotive] = false_discovery_rate(
                            tp=tp, fp=fp)
                        test_step_info[val_type]['TNR'][k][emotive] = true_negative_rate(
                            tn=tn, fp=fp)
                        test_step_info[val_type]['TPR'][k][emotive] = true_positive_rate(
                            tp=tp, fn=fn)
                        test_step_info[val_type]['NPV'][k][emotive] = negative_predictive_value(
                            tn=tn, fn=fn)
                        test_step_info[val_type]['FNR'][k][emotive] = false_negative_rate(
                            fn=fn, tp=tp)
                        test_step_info[val_type]['FOR'][k][emotive] = false_omission_rate(
                            fn=fn, tn=tn)
                        test_step_info[val_type]['LR+'][k][emotive] = positive_likelihood_ratio(
                            tp=tp, fp=fp, tn=tn, fn=fn)
                        test_step_info[val_type]['LR-'][k][emotive] = negative_likelihood_ratio(
                            tp=tp, fp=fp, tn=tn, fn=fn)
                        test_step_info[val_type]['PT'][k][emotive] = prevalence_threshold(
                            tp=tp, fp=fp, tn=tn, fn=fn)
                        test_step_info[val_type]['TS'][k][emotive] = threat_score(
                            tp=tp, fp=fp, fn=fn)

                for val_type in [actual_value_type, 'overall']:
                    # Update response percentage and unknown percentage
                    test_step_info[val_type]['response_percentage'][k][emotive] = (
                        test_step_info[val_type]['response_counts'][k][emotive] / (test_step_info[val_type]['testing_counter'][k][emotive])) * 100
                    test_step_info[val_type]['unknown_percentage'][k][emotive] = 100 - \
                        test_step_info[val_type]['response_percentage'][k][emotive]

    elif test_type == 'emotives_value':

        for emotive in test_step_info['actual'].keys():
            for k in test_step_info['predicted'].keys():
                if emotive in test_step_info['predicted'][k].keys():
                    test_step_info['residuals'][k][emotive] = compute_residual(actual=test_step_info['actual'][emotive],
                                                                               predicted=test_step_info['predicted'][k][emotive])
                    test_step_info['abs_residuals'][k][emotive] = compute_abs_residual(actual=test_step_info['actual'][emotive],
                                                                                       predicted=test_step_info['predicted'][k][emotive])
                    test_step_info['squared_residuals'][k][emotive] = compute_squared_residual(actual=test_step_info['actual'][emotive],
                                                                                               predicted=test_step_info['predicted'][k][emotive])

                    previous_rmse = test_step_info['rmse'][k][emotive]
                    previous_smape = test_step_info['smape'][k][emotive]
                    # touch 1-smape
                    test_step_info['1-smape'][k][emotive]
                    if test_step_info['predicted'][k][emotive] == np.nan:
                        continue

                    count = test_step_info['response_counts'][k][emotive]
                    # compute new_rmse
                    test_step_info['smape'][k][emotive] = smape(previous_smape=previous_smape,
                                                                count=count,
                                                                abs_residual=test_step_info['abs_residuals'][k][emotive],
                                                                predicted=test_step_info['predicted'][k][emotive],
                                                                actual=test_step_info['actual'][emotive])
                    test_step_info['rmse'][k][emotive] = rmse(previous_rmse=previous_rmse,
                                                              count=count,
                                                              squared_residual=test_step_info['squared_residuals'][k][emotive])
                    test_step_info['1-smape'][k][emotive] = 100 - \
                        (test_step_info['smape'][k][emotive])
                    test_step_info['response_counts'][k][emotive] += 1

                # Update response percentage and unknown percentage
                test_step_info['response_percentage'][k][emotive] = (
                    test_step_info['response_counts'][k][emotive] / (test_step_info['testing_counter'][emotive])) * 100
                test_step_info['unknown_percentage'][k][emotive] = 100 - \
                    test_step_info['response_percentage'][k][emotive]

    return test_step_info



import time
import random
import numpy as np
import os
import traceback
import logging
from typing import Tuple, Any, Union
from collections.abc import Callable
from copy import deepcopy
import multiprocessing
from deap import base, creator, tools, algorithms
from deap.tools import Logbook
from tqdm.auto import tqdm
from ia.gaius.agent_client import AgentClient
from ia.gaius.manager import AgentManager
from ia.gaius.pvt import PerformanceValidationTest
from hashlib import sha1

logger = logging.getLogger(__name__)

# define maximum permissible values for performing optimization
# on several genes. Restricts range of values explored by optimization
MAX_PRED_LIMIT = 100
MAX_RT_LIMIT = 0.999

SUPPORTED_GENES = ['recall_threshold', 'max_predictions']
SPAWN_AGENT_LOCK = multiprocessing.Lock()

def generate_gene_data(config):

    result = {}
    for gene, params in config.items():
        if params["step"] == 0:
            value = round(random.uniform(params["start"], params["stop"]), 2)
        else:
            values = range(params["start"], params["stop"], params["step"])
            # print(gene, list(values))
            value = random.choice(values)

        result[gene] = value
    return result

class GenomeOptimizer:

    def __init__(self,
                 path_to_original_genome: str,
                 nodes_to_optimize: list,
                 pvt_config: dict,
                 gene_config: dict, 
                 evolutionary_params: dict,
                 agent_constructor=AgentClient,
                 agent_kwargs: dict = None,
                 pvt_constructor=PerformanceValidationTest,
                 weights: dict = {"accuracy": 1.0, "precision": 1.0},
                 ):
        self.am = AgentManager(local=False)
        self.genome_path = path_to_original_genome
        self.nodes_to_optimize = nodes_to_optimize
        self.agent_constructor = agent_constructor
        self.agent_kwargs = agent_kwargs
        self.pvt_config = pvt_config
        self.gene_config = gene_config
        self.pvt_constructor = pvt_constructor
        self._results = None  # internal variable to hold copy of results

        # Parameters for running evolution
        self.evolutionary_params = evolutionary_params
        self.fitness_weight_dict = weights

        try:
            del creator.FitnessMax
            del creator.Individual
        except:
            pass
        # Create and register DEAP types
        creator.create("FitnessMax", base.Fitness,
                       weights=(self.fitness_weight_dict["accuracy"],
                                self.fitness_weight_dict["precision"]))
        creator.create("Individual", dict, fitness=creator.FitnessMax)

        self.stats = tools.Statistics(key=lambda ind: ind)
        self.stats.register("average_fitness",
                            compute_metric_by_key, func=np.average)
        self.stats.register("std_dev_fitness",
                            compute_metric_by_key, func=np.std)
        self.stats.register("min_fitness", compute_metric_by_key, func=np.min)
        self.stats.register("max_fitness", compute_metric_by_key, func=np.max)

        # Create the toolbox
        self.toolbox = base.Toolbox()
        self.toolbox.register(
            'gen_individual', self.generate_individual, nodes_to_optimize=self.nodes_to_optimize)
        # self.toolbox.register("individual", creator.Individual)
        self.toolbox.register("population", tools.initRepeat,
                              list, self.toolbox.gen_individual)

        # Register evolutionary operators
        self.toolbox.register("evaluate", evaluate, am=self.am, agent_constructor=self.agent_constructor,
                              agent_kwargs=self.agent_kwargs,
                              pvt_config=self.pvt_config, genome_path=self.genome_path,
                              pvt_constructor=self.pvt_constructor,
                              nodes_to_optimize=nodes_to_optimize)
        self.toolbox.register("mate", crossover, toolbox=self.toolbox, gene_config=self.gene_config)
        self.toolbox.register("mutate", mutate, toolbox=self.toolbox, gene_config=self.gene_config)
        self.toolbox.register("select", tools.selTournament, tournsize=3)

    def generate_individual(self, nodes_to_optimize: list):

        # return creator.Individual({node: {"recall_threshold":  round(random.uniform(0.001, 0.999), 4),
        #                            "max_predictions": round(random.uniform(1, 100))} for node in nodes_to_optimize})
        return creator.Individual({node: generate_gene_data(self.gene_config) for node in nodes_to_optimize})

    def evolve(self, pool=None):
        if pool:
            self.toolbox.register("map", pool.map)
        params = deepcopy(self.evolutionary_params)
        population = self.toolbox.population(params.pop("npop"))

        self._results = algorithms.eaSimple(
            population=population, toolbox=self.toolbox, stats=self.stats, **params)
        return deepcopy(self._results)

    def multiprocessed_evolve(self, n_proc: int = None, start_hoster: bool=True) -> Tuple[Any, Logbook]:
        """Driver function to run genetic optimization using multithreading.
        Utilizes AgentManager to spawn and manage multiple agents

        Args:
            n_proc (int, optional): Number of processing cores to use. Defaults to minimum of (core_count, population_size).

        Returns:
            Tuple[Any, Logbook]
        """

        pop = None
        # set max number of processes to minimum of population size, core count
        if n_proc == None:
            CPU_COUNT = os.cpu_count()
            if CPU_COUNT == None:
                CPU_COUNT = 1

            n_proc = min([self.evolutionary_params['npop'], CPU_COUNT])
        try:
            SPAWN_AGENT_LOCK.release() # ensure lock is released before starting multiprocessing pool
        except:
            pass

        try:
            with multiprocessing.Pool(processes=n_proc) as pool:
                if start_hoster:
                    self.am.start_hoster()
                pop = self.evolve(pool=pool)
        except Exception as e:
            traceback.print_exc()
            logger.exception("Exception during multiprocessed_evolve")
            logger.info("Killing started agents")
            self.am.kill_all_agents()
            pass

        return pop


def compute_metric_by_key(pop: list, func: Callable):
    for ind in pop:
        ind.fitness_overall = np.average(ind.fitness.values, weights=[
                                         int(weight) for weight in ind.fitness.weights])
    return func([ind.fitness_overall for ind in pop])


def evaluate(individual,
             am: AgentManager,
             agent_constructor: AgentClient,
             agent_kwargs,
             pvt_config: dict,
             genome_path: str,
             pvt_constructor: Callable,
             nodes_to_optimize: list) -> Tuple[float, float]:
    """Driver function for conducting PVT on each individual

    Args:
        individual: The individual (including GCPs)
        am (AgentManager): AgentManager used to spawn a new agent
        pvt_config (dict): Configuration for PVT test
        genome_path (str): Genome used to spawn agent
        pvt_constructor (PerformanceValidationTest): PVT class
        nodes_to_optimize (list): list of nodes to optimize

    Returns:
        tuple: accuracy, precision
    """
    am.remediate_dead_agents()
    # Copy baseline config to assign agent
    test_config = deepcopy(pvt_config)
    if agent_kwargs == None:
        agent_kwargs = {}
    # logger.debug(f"{individual = }")
    ind_copy = deepcopy(individual)
    agent_name = sha1(str(ind_copy).encode(
        'utf-8', 'replace')).hexdigest()[:4] + "-" + multiprocessing.current_process().name

    accuracy = 0.0
    precision = 0.0

    # check if agent already alive
    if agent_name in am.current_agents:
        am.delete_agent(agent_name)
    released = True
    try:
        SPAWN_AGENT_LOCK.acquire()
        released = False

        with am.agent_context(genome_file=genome_path, user_id='optimizer', agent_id=agent_name, agent_name=agent_name) as agent:
            SPAWN_AGENT_LOCK.release()
            released = True

            time.sleep(5)
            agent = agent_constructor(agent._bottle_info, **agent_kwargs)
            agent.connect()
            logger.debug(f'in evaluate, {individual = }')
            for node in nodes_to_optimize:
                agent.change_genes(dict(individual[node]), nodes=node)

            test_config["agent"] = agent
            pvt = pvt_constructor(**test_config)
            agent.set_summarize_for_single_node(False)

            pvt.conduct_pvt()

            accuracy = pvt.testing_log[-1][-1]['overall_metrics']["accuracy"]["hive"]
            precision = pvt.testing_log[-1][-1]['overall_metrics']["precision"]["hive"]
            logger.info(
                f"Individual {individual} results: {accuracy = },{precision = }")

            SPAWN_AGENT_LOCK.acquire()
            released = False
    except:
        logger.exception("exception in evaluate thread")
    finally:
        if not released:
            SPAWN_AGENT_LOCK.release()
    return accuracy, precision


def crossover(individual1, individual2, toolbox, gene_config):

    child1, child2 = [dict(toolbox.clone(ind))
                      for ind in (individual1, individual2)]

    for node in individual1.keys():
        for key in individual1[node].keys():
            parent_values = [individual1[node][key], individual2[node][key]]
            quartiles = np.percentile(parent_values, [25, 75])
            
            # if variable is continuous, round to 4 decimals
            # otherwise, round to integer value
            if gene_config[key]['step'] == 0: 
                child1[node][key] = round(quartiles[0], 4)
                child2[node][key] = round(quartiles[1], 4)
            else:
                child1[node][key] = round(quartiles[0])
                child2[node][key] = round(quartiles[1])

            # clip values to range provided in gene_config
            child1[node][key] = min(child1[node][key], gene_config[key]['stop'])
            child1[node][key] = max(child1[node][key], gene_config[key]['start'])
            
            child2[node][key] = min(child2[node][key], gene_config[key]['stop'])
            child2[node][key] = max(child2[node][key], gene_config[key]['start'])
            
    return creator.Individual(child1), creator.Individual(child2)


def mutate(individual: dict, toolbox, gene_config, mu: float = 0.0, sigma: float = 0.1):
    """Mutate an individual by applying an offset sampled from a gaussian distribution
    centered at mu, with std deviation sigma

    Args:
        individual (dict): starting point
        toolbox (_type_): toolbox
        mu (float, optional): center of distribution. Defaults to 0.0.
        sigma (float, optional): standard deviation of distribution. Defaults to 0.1.

    Returns:
        creator.Individual: the mutated individual
    """
    new_individual = toolbox.gen_individual()
    # del new_individual.fitness.values

    # Mutate a float individual by adding Gaussian noise
    for node in individual:
        for k in individual[node]:
            logger.debug(
                f'mutating {k} from node({node}) on individual {individual}')
            new_individual[node][k] = mutate_variable(
                val=individual[node][k], key=k, mu=mu, sigma=sigma, gene_config=gene_config)

    return creator.Individual(new_individual),


def mutate_variable(val: Union[float, int], key: str, mu: float, sigma: float, gene_config: dict) -> float:
    """Mutate a variable on an individual, using gaussian noise. Performs bound checking on variable

    Args:
        val (float): initial value
        key (str): variable name
        mu (int, optional): center of gaussian noise. Defaults to 0.
        sigma (float, optional): percent deviation. Will be scaled by the maximum allowed value for each variable. Defaults to 0.1 (e.g. 10%).

    Returns:
        Union[float, int]: new value
    """
    # if key == 'recall_threshold':
    #     val = val + random.gauss(mu=mu, sigma=sigma*MAX_RT_LIMIT)
    #     val = round(max(min(val, 0.9999), 0.0001), 4)
    # elif key == 'max_predictions':
    #     val = val + random.gauss(mu=mu, sigma=sigma*MAX_PRED_LIMIT)
    #     val = round(max(min(val, MAX_PRED_LIMIT), 1))
    # return val

    # Check if continous variable, and round to integer if it is
    trunc = 4 if gene_config[key] == 0 else None

    upper_bound = gene_config[key]["stop"]
    lower_bound = gene_config[key]["start"]
    val = val + random.gauss(mu=mu, sigma=sigma*upper_bound)

    val = round(val, trunc)
    val = max(min(val, upper_bound), lower_bound)
    
    return val





"""Utilities for PVT computations"""
from sklearn.metrics import confusion_matrix, ConfusionMatrixDisplay
import matplotlib.pyplot as plt
from numpy import nan, sign, isnan
import pandas as pd
import plotly.graph_objects as go
from copy import deepcopy
import logging
import math
from pathlib import Path
from ia.gaius.agent_client import AgentClient
from ia.gaius.prediction_models import most_common_ensemble_model_classification, prediction_ensemble_modeled_emotives

logger = logging.getLogger(__name__)

def init_emotive_on_node(emotive: str, node: str, test_step_info: dict):
    """Helper function to initialize emotive information for live messages.
    Used if new emotive is encountered during testing
    (emotive only seen in specific records, not consistently across all)

    Args:
        emotive (str): emotive name
        node (str): node to initialize emotive on
        test_step_info (dict): dictionary of live information, which should
            be initialized with new emotive
    """
    test_step_info['overall']['response_counts'][node][emotive] = 0
    test_step_info['overall']['true_positive'][node][emotive] = 0
    test_step_info['overall']['true_negative'][node][emotive] = 0
    test_step_info['overall']['false_positive'][node][emotive] = 0
    test_step_info['overall']['false_negative'][node][emotive] = 0
    test_step_info['overall']['testing_counter'][node][emotive] = 0

    # init positive and negative metrics
    test_step_info['positive']['response_counts'][node][emotive] = 0
    test_step_info['positive']['true_positive'][node][emotive] = 0
    test_step_info['positive']['true_negative'][node][emotive] = 0
    test_step_info['positive']['false_positive'][node][emotive] = 0
    test_step_info['positive']['false_negative'][node][emotive] = 0
    test_step_info['positive']['testing_counter'][node][emotive] = 0

    test_step_info['negative']['response_counts'][node][emotive] = 0
    test_step_info['negative']['true_positive'][node][emotive] = 0
    test_step_info['negative']['true_negative'][node][emotive] = 0
    test_step_info['negative']['false_positive'][node][emotive] = 0
    test_step_info['negative']['false_negative'][node][emotive] = 0
    test_step_info['negative']['testing_counter'][node][emotive] = 0
    return


def init_emotive_polarity_results():
    result_dict = {'overall': {},
                   'positive': {},
                   'negative': {}}

    for key, value in result_dict.items():
        value['true_positive'] = 0
        value['false_positive'] = 0
        value['true_negative'] = 0
        value['false_negative'] = 0
        value['unknown_percentage'] = 0.0
        value['response_percentage'] = 0.0
        value['response_counts'] = 0
        value['accuracy'] = 0.0
        value['precision'] = 0.0
        value['training_counter'] = 0
        value['testing_counter'] = 0
        value['counter'] = 0
        value['FPR'] = 0.0
        value['FDR'] = 0.0
        value['TNR'] = 0.0
        value['TPR'] = 0.0
        value['NPV'] = 0.0
        value['FNR'] = 0.0
        value['FOR'] = 0.0
        value['LR+'] = 0.0
        value['LR-'] = 0.0
        value['PT'] = 0.0
        value['TS'] = 0.0

    return result_dict


def check_answer_correctness(predicted, actual, test_type: str) -> bool:
    if test_type == 'classification':
        # predicted should be a string, actual should be a list
        return predicted in actual
    elif test_type == 'emotives_polarity':
        # predicted and actual should be numbers
        predicted_sign = sign(predicted)
        actual_sign = sign(actual)

        return predicted_sign == actual_sign
    else:
        raise Exception(
            f"Check answer correctness received invalid test_type {test_type}")


def on_error_learn(agent: AgentClient, test_type: str, query_nodes: list, predictions: dict, current_labels: list, record_emotive_set: dict):
    correct_dict = {node: False for node in query_nodes}
    if test_type == 'classification':
        for node in correct_dict:
            correct_dict[node] = check_answer_correctness(predicted=most_common_ensemble_model_classification(
                predictions[node]), actual=current_labels, test_type='classification')
        pass
    elif test_type == 'emotives_polarity':
        for node in correct_dict:
            ensemble_emotives = prediction_ensemble_modeled_emotives(
                predictions[node])
            correct_dict[node] = all([check_answer_correctness(predicted=ensemble_emotives.get(
                emotive, 0.0), actual=emotive_value, test_type='emotives_polarity') for emotive, emotive_value in record_emotive_set.items()])

        pass
    else:
        raise Exception(
            """on_error learning strategy only permitted for classification and emotives_polarity
                        """
        )
    for node, answer_correct in correct_dict.items():
        if not answer_correct:
            agent.learn(nodes=[node])


def compute_residual(predicted: float, actual: float) -> float:
    """Compute residual given predicted and actual

    Args:
        predicted (float): predicted emotive value
        actual (float): actual emotive value
    """

    return (actual - predicted)


def compute_abs_residual(predicted: float, actual: float):
    """Compute absolute residual

    Args:
        predicted (float): predicted emotive value
        actual (float): actual emotive value
    """
    return abs(actual - predicted)


def compute_squared_residual(predicted: float, actual: float):
    """Compute absolute residual

    Args:
        predicted (float): predicted emotive value
        actual (float): actual emotive value
    """
    return math.pow(actual - predicted, 2)


def smape(previous_smape: float, count: int, abs_residual: float, predicted: float, actual: float):
    """Computes the new SMAPE, given previous smape,count, predicted and actual value

    Args:
        count (int): Response count for specific emotive on node
        predicted (float): predicted emotive value
        actual (float): actual emotive value
        previous_smape (float): Previous smape.
    """
    return 100 * ((((previous_smape / 100) * count) + ((2 * abs_residual) / (abs(predicted) + abs(actual)))) / (count + 1))


def rmse(previous_rmse: float, count: int, squared_residual: float):
    """Compute new RMSE given previous RMSE value, count, and new squared residual

    Args:
        previous_rmse (float): previous RMSE value
        count (int): Response count for specific emotive on node
        squared_residual (float): current squared residual value
    """
    return math.sqrt(((math.pow(previous_rmse, 2) * count) + squared_residual) / (count + 1))


def f1_score(tp: int, fp: int, fn: int):
    """Compute F1 Score

    Args:
        tp (int): True Positive count
        fp (int): False Positive count
        fn (int): False Negative count
    """
    f1 = 0.0
    try:
        f1 = (2 * tp) / ((2 * tp) + fp + fn)
    except ZeroDivisionError:
        pass
    return f1


def false_discovery_rate(tp: int, fp: int):
    """Compute FDR

    Args:
        tp (int): True Positive count
        fp (int): False Positive count
    """
    fdr = 0.0
    try:
        fdr = 100.0 * fp / (fp + tp)
    except ZeroDivisionError:
        pass
    return fdr


def true_negative_rate(tn: int, fp: int):
    """Compute FDR

    Args:
        tn (int): True Negative count
        fp (int): False Positive count
    """
    tnr = 0.0
    try:
        tnr = 100.0 * tn / (tn + fp)
    except ZeroDivisionError:
        pass
    return tnr


def true_positive_rate(tp: int, fn: int):
    """Compute FDR

    Args:
        tp (int): True Positive count
        fn (int): False Negative count
    """
    tpr = 0.0
    try:
        tpr = 100.0 * tp / (tp + fn)
    except ZeroDivisionError:
        pass
    return tpr


def negative_predictive_value(tn: int, fn: int):
    """Compute NPV

    Args:
        tn (int): True Negative count
        fn (int): False Negative count
    """
    npv = 0.0
    try:
        npv = 100.0 * tn / (tn + fn)
    except ZeroDivisionError:
        pass
    return npv


def false_negative_rate(fn: int, tp: int):
    """Compute FNR

    Args:
        fn (int): False Negative count
        tp (int): True Positive count
    """
    fnr = 0.0
    try:
        fnr = 100.0 * fn / (fn + tp)
    except ZeroDivisionError:
        pass
    return fnr


def false_omission_rate(fn: int, tn: int):
    """Compute FOR

    Args:
        fn (int): False Negative count
        tn (int): True Negative count
    """
    false_or = 0.0
    try:
        false_or = 100.0 * fn / (fn + tn)
    except ZeroDivisionError:
        pass
    return false_or


def false_positive_rate(fp: int, tn: int):
    """Compute FPR

    Args:
        fp (int): False Positive count
        tn (int): True Negative count
    """
    fpr = 0.0
    try:
        fpr = 100.0 * fp / (fp + tn)
    except ZeroDivisionError:
        pass
    return fpr


def positive_likelihood_ratio(tp: int, fp: int, tn: int, fn: int):
    """Compute LR+

    Args:
        tp (int): True Positive count
        fp (int): False Positive count
        tn (int): True Negative count
        fn (int): False Negative count
    """
    lr_plus = 0.0
    try:
        lr_plus = true_positive_rate(
            tp=tp, fn=fn) / false_positive_rate(fp=fp, tn=tn)
    except ZeroDivisionError:
        pass
    return lr_plus


def negative_likelihood_ratio(tp: int, fp: int, tn: int, fn: int):
    """Compute LR-

    Args:
        tp (int): True Positive count
        fp (int): False Positive count
        tn (int): True Negative count
        fn (int): False Negative count
    """
    lr_minus = 0.0
    try:
        lr_minus = false_negative_rate(
            fn=fn, tp=tp) / true_negative_rate(tn=tn, fp=fp)
    except ZeroDivisionError:
        pass
    return lr_minus


def prevalence_threshold(tp: int, fp: int, tn: int, fn: int):
    """Compute PT

    Args:
        tp (int): True Positive count
        fp (int): False Positive count
        tn (int): True Negative count
        fn (int): False Negative count
    """
    pt = 0.0
    try:
        pt = math.sqrt(false_positive_rate(fp=fp, tn=tn)) / (math.sqrt(
            true_positive_rate(tp=tp, fn=fn)) + math.sqrt(false_positive_rate(fp=fp, tn=tn)))
    except ZeroDivisionError:
        pass
    return pt


def threat_score(tp: int, fp: int, fn: int):
    """Compute TS

    Args:
        tp (int): True Positive count
        fp (int): False Positive count
        fn (int): False Negative count
    """
    ts = 0.0
    try:
        ts = 100.0 * tp / (tp + fn + fp)
    except ZeroDivisionError:
        pass
    return ts


def update_accuracy(tp: int, tn: int, overall_count: int) -> float:
    """Update accuracy metrics

    Args:
        tp (int): True Positives
        tn (int): True Negatives
        overall_count (int): current testing record count

    Returns:
        float: accuracy
    """
    accuracy = 0.0
    try:
        accuracy = 100.0 * (tp + tn) / overall_count
    except ZeroDivisionError:
        pass
    return accuracy


def update_precision(tp: int, tn: int, response_count: int) -> float:
    """Update precision metrics

    Args:
        tp (int): True Positives
        tn (int): True Negatives
        response_count (int): node response count

    Returns:
        float: precision
    """
    precision = 0.0
    try:
        precision = 100.0 * (tp + tn) / response_count
    except ZeroDivisionError:
        pass
    return precision


def plot_confusion_matrix(test_num: int, class_metrics_data_structures: dict, results_dir=None):
    """
    Takes a node classification test to create a confusion matrix.
    This version includes the i_dont_know or unknown label.
    """

    for node_name, class_metrics_data in class_metrics_data_structures.items():
        # print(f'-------------Test#{test_num}-{node_name}-Plots-------------')
        sorted_labels = deepcopy(class_metrics_data['labels'])
        label_set = set(sorted_labels).union(class_metrics_data['predictions'])
        
        sorted_labels = [str(label)
                         for label in label_set if label is not None]
        sorted_labels.append(str(None))
        # sorted_labels = sorted(class_metrics_data['labels'])
        actuals = [[str(elem) for elem in act]
                   for act in class_metrics_data['actuals']]
        preds = [str(pred) for pred in class_metrics_data['predictions']]
        try:
            cm = confusion_matrix(actuals,  # TODO: each "actual" is a list, to support multiclass labels. Need to find solution here
                                  preds,
                                  labels=sorted_labels)
        except ValueError as e:  # thrown when there are no predictions, and nothing to plot
            print(str(e))
            logger.exception("broke in plot_confusion_matrix")
            return
        disp = ConfusionMatrixDisplay(confusion_matrix=cm,
                                      display_labels=sorted_labels)

        disp.plot()
        disp.ax_.set_title(f'Test#{test_num}-{node_name} Confusion Matrix')
        current_figure = plt.gcf()
        if is_notebook():
            plt.show()
        if results_dir is not None:
            cf_filename = Path(results_dir).joinpath(
                f'./confusion_matrix_{node_name}_test_{test_num}.png')
            # cf_filename = f'./confusion_matrix_{node_name}_test_{test_num}.png'
            # print(f'attempting to save confusion_matrix to: {cf_filename}')
            try:
                current_figure.savefig(cf_filename)
            except Exception as e:
                print(
                    f'error saving confusion matrix to {cf_filename}: {str(e)}')
        plt.close()

def plot_emotives_value_charts(test_num: int, emotive_value_results: dict, results_filepath: str, QUIET: bool = False):
    """Plot charts for emotives value charts
    """
    for node_name, node_emotive_metrics in emotive_value_results.items():
        logger.debug(
            f'-----------------Test#{test_num}-{node_name}-Plots-----------------')
        for emotive_name, data in sorted(node_emotive_metrics.items()):
            labels = 'precision', 'miss'
            if isnan(data['metrics']['1-smape']):
                continue
            if data['metrics']['1-smape'] is None:
                sizes = [0, 100]
            else:
                sizes = [data['metrics']['1-smape'],
                            100 - data['metrics']['1-smape']]
            explode = (0, 0)
            _, ax1 = plt.subplots()
            ax1.title.set_text(f'{node_name} - {emotive_name}')
            ax1.pie(sizes, explode=explode, labels=labels, autopct='%1.1f%%',
                    shadow=True, startangle=90)
            # Equal aspect ratio ensures that pie is drawn as a circle.
            ax1.axis('equal')
            colors = ['gray', 'skyblue']
            patches, texts = plt.pie(sizes, colors=colors, startangle=90)
            plt.legend(patches, labels, loc="best")
            supplementary_data = {'SMAPE Precision': data['metrics']['1-smape'],
                                    'RMSE': data['metrics']['rmse']}
            plt.figtext(0, 0, f"{pd.Series(supplementary_data).round(1).to_string()}", ha="center", fontsize=18, bbox={
                        "facecolor": "orange", "alpha": 0.5, "pad": 5})
            pie_plot_filepath = Path(f"{results_filepath}").joinpath(
                f"./test_{test_num}_{node_name}_{emotive_name}_pie.png")
            pva_plot_filepath = Path(f"{results_filepath}").joinpath(
                f"./test_{test_num}_{node_name}_{emotive_name}_pva.png")

            df = pd.DataFrame({'predicted': data['predictions'],
                                'actuals': data['actuals']})
            predicted_vs_actual = go.Figure()
            predicted_vs_actual.update_layout(title=dict(text=f"Test #{test_num}: {emotive_name} values on {node_name}"),
                                                xaxis=dict(
                                                    title="Testing Record"),
                                                yaxis=dict(
                                                    title=f"{emotive_name} Value")
                                                )
            predicted_vs_actual.add_trace(go.Scatter(
                x=df.index, y=df.actuals, fill='tozeroy', name='Actual'))
            predicted_vs_actual.add_trace(go.Scatter(
                x=df.index, y=df.predicted, fill='tonexty', name='Predicted'))
            if results_filepath is not None:
                try:
                    plt.savefig(pie_plot_filepath, dpi=300,
                                bbox_inches='tight')
                    predicted_vs_actual.write_image(file=pva_plot_filepath)
                except Exception as e:
                    logger.debug(
                        f"Not able to save figure in assigned results directory! Please add an appropriate directory: {str(e)}")
                    pass
            if is_notebook() and not QUIET:
                plt.show()
                predicted_vs_actual.show()

    plt.close('all')

def is_notebook() -> bool:  # pragma: no cover (helper function to determine if we are in a jupyter notebook)
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False      # Probably standard Python interpreter



from ia.gaius.data_language import CLEAR_WM, LEARN
# from ia.gaius.pvt.pvt_utils import make_modeled_emotives_
from ia.gaius.prediction_models import prediction_ensemble_modeled_emotives
class Tester:

    def __init__(self, **kwargs):
        self.bottle = kwargs['bottle']
        self.utype = kwargs['utype']  # either 'value' or 'polarity'
        self.learning_strategy = kwargs['learning_strategy']

    def next_test_prep(self):
        """Anything to reset in between multiple test runs."""
        return

    def train(self, sequence):
        record = {'phase': 'training', 'historical_expecting': sum([sum(data['emotives'].values()) for data in sequence])}
        self.bottle.clear_wm()
        for data in sequence:
            self.bottle.observe(data)
        self.bottle.learn()
        return record

    def test(self, sequence):
        record = {'phase': 'testing'}
        historical_expecting = sum([sum(data['emotives'].values()) for data in sequence])
        record['historical_expecting'] = historical_expecting
        self.bottle.clear_wm()
        for data in sequence:
            self.bottle.observe(data)

        record['node_predictions'] = {}
        predicton_error = True  # If any of the nodes are right, then don't bother learning if on_error learning strategy is employed.
        hive_prediction = []
        answer = self.bottle.get_predictions()
        # HERE IS WHERE WE MODEL OUR PREDICTIONS #######
        utility_preds = []
        for node, ensemble in answer.items():
            predicted_value = prediction_ensemble_modeled_emotives(ensemble)
            if isinstance(predicted_value, dict) and len(predicted_value) != 0:
                predicted_value = sum(list(predicted_value.values()))/ len(predicted_value)
            else:
                predicted_value = 0
            utility_preds.append(predicted_value)
            record['node_predictions'][node] = predicted_value

        hive_prediction = sum(utility_preds)/len(utility_preds)
        record['node_predictions']['hive'] = hive_prediction
        if hive_prediction == historical_expecting:
            predicton_error = False

        if self.learning_strategy == 'continuous' or (
                self.learning_strategy == 'on-error' and predicton_error):
            # NOTE!: Unlike regular symbols for the classification tests, utilities pass through
            # all nodes, so we don't need to send directly to egress nodes.  Simply send through ingress nodes.
            data = {"strings": [], "vectors": [], "emotives": {'utility': historical_expecting}}
            self.bottle.observe(data)
            self.bottle.learn()
        return record
